#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import os
import sys
import time
import subprocess
import json
import signal
from weakref import proxy

import psutil

try:
    import psychopy.logging as psycho_logging
except ImportError:
    psycho_logging = None
from ..lazy_import import lazy_import
from .. import IOHUB_DIRECTORY
from ..util import yload, yLoader
from ..errors import print2err, ioHubError, printExceptionDetailsToStdErr
from ..util import isIterable, updateDict, win32MessagePump
from ..devices import DeviceEvent, import_device
from ..devices.computer import Computer
from ..devices.experiment import MessageEvent, LogEvent
from ..constants import DeviceConstants, EventConstants
from psychopy import constants

getTime = Computer.getTime

_currentSessionInfo = None

def windowInfoDict(win):
    windict = dict(handle=win._hw_handle, pos=win.pos, size=win.size,
                   units=win.units, useRetina=win.useRetina, monitor=None)
    if win.monitor:
        windict['monitor'] = dict(resolution=win.monitor.getSizePix(),
                                  width=win.monitor.getWidth(),
                                  distance=win.monitor.getDistance())
    return windict

def getFullClassName(klass):
    module = klass.__module__
    if module == 'builtins':
        return klass.__qualname__  # avoid outputs like 'builtins.str'
    return module + '.' + klass.__qualname__

class DeviceRPC():
    '''
    ioHubDeviceView creates an RPC interface with the iohub server. Each
    iohub device method exposed by an ioHubDeviceView is represented
    by an associated DeviceRPC instance.
    '''
    _log_time_index = DeviceEvent.EVENT_HUB_TIME_INDEX
    _log_text_index = LogEvent.CLASS_ATTRIBUTE_NAMES.index('text')
    _log_level_index = LogEvent.CLASS_ATTRIBUTE_NAMES.index('log_level')

    def __init__(self, sendToHub, device_class, method_name):
        self.device_class = device_class
        self.method_name = method_name
        self.sendToHub = sendToHub

    @staticmethod
    def _returnarg(a): # pragma: no cover
        return a

    def __call__(self, *args, **kwargs):
        # Send the device method call request to the ioHub Server and wait
        # for the method return value sent back from the ioHub Server.
        r = self.sendToHub(('EXP_DEVICE', 'DEV_RPC', self.device_class,
                            self.method_name, args, kwargs))
        
        if r is None:
            # print("r is None:",('EXP_DEVICE', 'DEV_RPC', self.device_class,
            #                 self.method_name, args, kwargs))
            return None
        
        r = r[1:]
        if len(r) == 1:
            r = r[0]

        if self.method_name != 'getEvents':
            return r

        # The result of a call to an iohub Device getEvents() method
        # gets some special handling, converting the returned events
        # into the desired object type, etc...
        asType = 'namedtuple'
        if 'asType' in kwargs:
            asType = kwargs['asType']
        elif 'as_type' in kwargs:
            asType = kwargs['as_type']

        conversionMethod = self._returnarg
        if asType == 'dict':
            conversionMethod = ioHubConnection.eventListToDict
        elif asType == 'object':
            conversionMethod = ioHubConnection.eventListToObject
        elif asType == 'namedtuple':
            conversionMethod = ioHubConnection.eventListToNamedTuple

        if self.device_class != 'Experiment':
            return [conversionMethod(el) for el in r]

        EVT_TYPE_IX = DeviceEvent.EVENT_TYPE_ID_INDEX
        LOG_EVT = LogEvent.EVENT_TYPE_ID
        toBeLogged = [el for el in r if el[EVT_TYPE_IX] == LOG_EVT]
        for l in toBeLogged:
            r.remove(l)
            if psycho_logging:
                ltime = l[self._log_time_index]
                ltext = l[self._log_text_index]
                llevel = l[self._log_level_index]
                psycho_logging.log(ltext, llevel, ltime)
        return [conversionMethod(el) for el in r]


# pylint: disable=protected-access

class ioHubDeviceView():
    """
    ioHubDeviceView is used as a client / PsychoPy process side representation
    of an ioHub device that is actually running on the separate iohub process.
    An ioHubDeviceView instance allows the PsychoPy script process to call
    public iohub device methods as if the device method calls were being made
    locally.

    The ioHubConnection class creates an ioHubDeviceView instance for each
    ioHub device being run during the experiment.

    ioHubDeviceView instances are never created directly by a user script,
    they are created for you by the ioHubConnection class when
    it connects to the ioHub Process.
    """

    def __init__(self, hubClient, device_class_path, device_class_name, device_config):
        self.hubClient = hubClient
        self.name = device_config.get('name', device_class_name.lower())
        self.device_class = device_class_name
        self.device_class_path=device_class_path

        rpc_request = ('EXP_DEVICE', 'GET_DEV_INTERFACE', device_class_name)
        r = self.hubClient._sendToHubServer(rpc_request)
        self._methods = r[1]

    def __getattr__(self, name):
        if name in self._methods:
            r = DeviceRPC(self.hubClient._sendToHubServer, self.device_class, name)
            return r
        raise AttributeError(self, name)

    def getName(self):
        """
        Gets the name given to the device in the ioHub configuration file.
        ( the device: name: property )

        Args:
            None

        Returns:
            (str): the user defined label / name of the device
        """
        return self.name

    def getIOHubDeviceClass(self, full=False):
        """
        Gets the ioHub Device class associated with the oHubDeviceView.
        This is specified for a device in the ioHub configuration file.
        ( the device: device_class: property )

        :param full:

        Returns:
            (class): ioHub Device class associated with this ioHubDeviceView

        """
        if full:
            return self.device_class_path
        return self.device_class

    def getDeviceInterface(self):
        """getDeviceInterface returns a list containing the names of all
        methods that are callable for the ioHubDeviceView object. Only public
        methods are included in the interface. Any method beginning with a
        '_' is not included.

        Args:
            None

        Returns:
            (tuple): list of method names in the ioHubDeviceView interface.

        """
        return self._methods

# pylint: enable=protected-access

class ioHubDevices():
    """
    Provides .name access to the the ioHub device's created when the ioHub
    Server is started. Each iohub device is accessible via a dynamically
    created attribute of this class, the name of which is defined by the
    device configuration 'name' setting. Each device attribute is an instance
    of the ioHubDeviceView class.

    A user script never creates an instance of this class directly, access
    is provided via the ioHubConnection.devices attribute.
    """

    def __init__(self, hubClient):
        self.hubClient = hubClient
        self._devicesByName = dict()

    def addDevice(self, name, d):
        setattr(self, name, d)
        self._devicesByName[name] = d

    def getDevice(self, name):
        return self._devicesByName.get(name)

    def getAll(self):
        return self._devicesByName.values()

    def getNames(self):
        return self._devicesByName.keys()

class ioHubConnection():
    """ioHubConnection is responsible for creating, sending requests to, and
    reading replies from the ioHub Process. This class is also used to
    shut down and disconnect the ioHub Server process.

    The ioHubConnection class is also used as the interface to any ioHub Device
    instances that have been created so that events from the device can be
    monitored. These device objects can be accessed via the ioHubConnection
    .devices attribute, providing 'dot name' access to enabled devices.
    Alternatively, the .getDevice(name) method can be used and will return
    None if the device name specified does not exist.

    Using the .devices attribute is handy if you know the name of the device
    to be accessed and you are sure it is actually enabled on the ioHub
    Process.

    An example of accessing a device using the .devices attribute::

        # get the Mouse device, named mouse
        mouse=hub.devices.mouse
        mouse_position = mouse.getPosition()

        print 'mouse position: ', mouse_position

        # Returns something like:
        # >> mouse position:  [-211.0, 371.0]
    """
    ACTIVE_CONNECTION = None

    def __init__(self, ioHubConfig=None, ioHubConfigAbsPath=None):
        if ioHubConfig:
            if not isinstance(ioHubConfig, dict):
                raise ioHubError(
                    'The provided ioHub Configuration is not a dictionary.',
                    ioHubConfig)

        if ioHubConnection.ACTIVE_CONNECTION is not None:
            raise RuntimeError('An existing ioHubConnection is already open.'
                                 ' Use ioHubConnection.getActiveConnection() '
                                 'to access it; or use ioHubConnection.quit() '
                                 'to close it.')
        Computer.psychopy_process = psutil.Process()

        # udp port setup
        self.udp_client = None

        # the dynamically generated object that contains an attribute for
        # each device registered for monitoring with the ioHub server so
        # that devices can be accessed experiment process side by device name.
        self.devices = ioHubDevices(self)

        # A circular buffer used to hold events retrieved from self.getEvents()
        # during self.wait() periods.
        self.allEvents = []

        self.experimentID = None
        self.experimentSessionID = None
        self._experimentMetaData = None
        self._sessionMetaData = None
        self._server_process = None
        self._iohub_server_config = None
        self._shutdown_attempted = False
        self._cv_order = None
        self._message_cache = []
        self.iohub_status = self._startServer(ioHubConfig, ioHubConfigAbsPath)
        if self.iohub_status != 'OK':
            raise RuntimeError('Error starting ioHub server: {}'.format(self.iohub_status))

    @classmethod
    def getActiveConnection(cls):
        return cls.ACTIVE_CONNECTION

    def getDevice(self, deviceName):
        """
        Returns the ioHubDeviceView that has a matching name (based on the
        device : name property specified in the ioHub_config.yaml for the
        experiment). If no device with the given name is found, None is
        returned. Example, accessing a Keyboard device that was named 'kb' ::

            keyboard = self.getDevice('kb')
            kb_events= keyboard.getEvent()

        This is the same as using the 'natural naming' approach supported
        by the .devices attribute, i.e::

            keyboard = self.devices.kb
            kb_events= keyboard.getEvent()

        However the advantage of using getDevice(device_name) is that an
        exception is not created if you provide an invalid device name,
        or if the device is not enabled on the ioHub server; None is returned
        instead.

        Args:
            deviceName (str): Name given to the ioHub Device to be returned

        Returns:
            The ioHubDeviceView instance for deviceName.
        """
        return self.devices.getDevice(deviceName)

    def getEvents(self, device_label=None, as_type='namedtuple'):
        """Retrieve any events that have been collected by the ioHub Process
        from monitored devices since the last call to getEvents() or
        clearEvents().

        By default all events for all monitored devices are returned,
        with each event being represented as a namedtuple of all event
        attributes.

        When events are retrieved from an event buffer, they are removed from
        that buffer as well.

        If events are only needed from one device instead of all devices,
        providing a valid device name as the device_label argument will
        result in only events from that device being returned.

        Events can be received in one of several object types by providing the
        optional as_type property to the method. Valid values for as_type are
        the following str values:

            * 'list': Each event is a list of ordered attributes.
            * 'namedtuple': Each event is converted to a namedtuple object.
            * 'dict': Each event converted to a dict object.
            * 'object': Each event is converted to a DeviceEvent subclass
                        based on the event's type.

        Args:
            device_label (str): Name of device to retrieve events for.
                                If None ( the default ) returns device events
                                from all devices.

            as_type (str): Returned event object type. Default: 'namedtuple'.

        Returns:
            tuple: List of event objects; object type controlled by 'as_type'.
        """
        r = None
        if device_label is None:
            events = self._sendToHubServer(('GET_EVENTS',))[1]
            if events is None:
                r = self.allEvents
            else:
                self.allEvents.extend(events)
                r = self.allEvents
            self.allEvents = []
        else:
            r = self.devices.getDevice(device_label).getEvents()

        if r:
            if as_type == 'list':
                return r

            conversionMethod = None
            if as_type == 'namedtuple':
                conversionMethod = self.eventListToNamedTuple
            elif as_type == 'dict':
                conversionMethod = self.eventListToDict
            elif as_type == 'object':
                conversionMethod = self.eventListToObject

            if conversionMethod:
                return [conversionMethod(el) for el in r]
            return r

        return []

    def clearEvents(self, device_label='all'):
        """Clears unread events from the ioHub Server's Event Buffer(s)
        so that unneeded events are not discarded.

        If device_label is 'all', ( the default ), then events from both the
        ioHub *Global Event Buffer* and all *Device Event Buffer's*
        are cleared.

        If device_label is None then all events in the ioHub
        *Global Event Buffer* are cleared, but the *Device Event Buffers*
        are unaffected.

        If device_label is a str giving a valid device name, then that
        *Device Event Buffer* is cleared, but the *Global Event Buffer* is not
        affected.

        Args:
            device_label (str): device name, 'all', or None

        Returns:
            None

        """
        if device_label.lower() == 'all':
            self.allEvents = []
            self._sendToHubServer(('RPC', 'clearEventBuffer', [True, ]))
            try:
                self.getDevice('keyboard')._clearLocalEvents()
            except:
                pass
        elif device_label in [None, '', False]:
            self.allEvents = []
            self._sendToHubServer(('RPC', 'clearEventBuffer', [False, ]))
            try:
                self.getDevice('keyboard')._clearLocalEvents()
            except:
                pass
        else:
            d = self.devices.getDevice(device_label)
            if d:
                d.clearEvents()

    def sendMessageEvent(self, text, category='', offset=0.0, sec_time=None):
        """
        Create and send an Experiment MessageEvent to the ioHub Server
        for storage in the ioDataStore hdf5 file.

        Args:
            text (str): The text message for the message event. 128 char max.

            category (str): A str grouping code for the message. Optional.
                            32 char max.

            offset (float): Optional sec.msec offset applied to the
                            message event time stamp. Default 0.

            sec_time (float): Absolute sec.msec time stamp for the message in.
                              If not provided, or None, then the MessageEvent
                              is time stamped when this method is called
                              using the global timer (core.getTime()).
        """
        self.cacheMessageEvent(text, category, offset, sec_time)
        self._sendToHubServer(('EXP_DEVICE', 'EVENT_TX', self._message_cache))
        self._message_cache = []

    def cacheMessageEvent(self, text, category='', offset=0.0, sec_time=None):
        """
        Create an Experiment MessageEvent and store in local cache.
        Message must be sent before it is saved to hdf5 file.

        Args:
            text (str): The text message for the message event. 128 char max.

            category (str): A str grouping code for the message. Optional.
                            32 char max.

            offset (float): Optional sec.msec offset applied to the
                            message event time stamp. Default 0.

            sec_time (float): Absolute sec.msec time stamp for the message in.
                              If not provided, or None, then the MessageEvent
                              is time stamped when this method is called
                              using the global timer (core.getTime()).
        """
        self._message_cache.append(MessageEvent._createAsList(text, # pylint: disable=protected-access
                                             category=category,
                                             msg_offset=offset,
                                             sec_time=sec_time))

    def sendMessageEvents(self, messageList=[]):
        if messageList:
            self.cacheMessageEvents(messageList)
        if self._message_cache:
            self._sendToHubServer(('EXP_DEVICE', 'EVENT_TX', self._message_cache))
            self._message_cache = []

    def cacheMessageEvents(self, messageList):
        for m in messageList:
            self._message_cache.append(MessageEvent._createAsList(**m))

    def getHubServerConfig(self):
        """Returns a dict containing the current ioHub Server configuration.

        Args:
            None

        Returns:
            dict: ioHub Server configuration.

        """
        return self._iohub_server_config

    def getSessionID(self):
        return self.experimentSessionID

    def getSessionMetaData(self):
        """Returns a dict representing the experiment session data that is
        being used for the current ioHub Experiment Session. Changing values in
        the dict has no effect on the session data that has already been saved
        to the ioHub DataStore.

        Args:
            None

        Returns:
            dict: Experiment Session metadata saved to the ioHub DataStore.
                  None if the ioHub DataStore is not enabled.
        """
        return self._sessionMetaData

    def getExperimentID(self):
        return self.experimentID

    def getExperimentMetaData(self):
        """Returns a dict representing the experiment data that is being used
        for the current ioHub Experiment.

        Args:
            None

        Returns:
            dict: Experiment metadata saved to the ioHub DataStore.
                  None if the ioHub DataStore is not enabled.

        """
        return self._experimentMetaData

    def wait(self, delay, check_hub_interval=0.02):
        # TODO: Integrate iohub event collection done in this version of wait
        # with psychopy wait() and deprecate this method.
        """Pause the experiment script execution for delay seconds.
        time.sleep() is used for delays > 0.02 sec (20 msec)

        During the wait period, events are received from iohub every
        'check_hub_interval' seconds, being buffered so they can be accessed
        after the wait duration. This is done for two reasons:

            * The iohub server's global and device level event buffers
              do not start to drop events if one of the (circular) event
              buffers becomes full during the wait duration.
            * The number of events in the iohub process event buffers does
              not becaome too large, which could result in a longer than
              normal getEvents() call time.

        Args:
            delay (float): The sec.msec delay until method returns.

            check_hub_interval (float): The sec.msec interval between calls to
                                        io.getEvents() during the delay period.
        Returns:
            float: The actual duration of the delay in sec.msec format.

        """
        stime = Computer.getTime()
        targetEndTime = stime + delay

        if check_hub_interval < 0:
            check_hub_interval = 0

        if check_hub_interval > 0:
            remainingSec = targetEndTime - Computer.getTime()
            while remainingSec > check_hub_interval+0.025:
                time.sleep(check_hub_interval)
                events = self.getEvents()
                if events:
                    self.allEvents.extend(events)
                # Call win32MessagePump so PsychoPy Windows do not become
                # 'unresponsive' if delay is long.
                win32MessagePump()
                remainingSec = targetEndTime - Computer.getTime()

        time.sleep(max(0.0, targetEndTime - Computer.getTime() - 0.02))

        while (targetEndTime - Computer.getTime()) > 0.0:
            pass

        return Computer.getTime() - stime

    def createTrialHandlerRecordTable(self, trials, cv_order=None):
        """
        Create a condition variable table in the ioHub data file based on
        the a psychopy TrialHandler. By doing so, the iohub data file
        can contain the DV and IV values used for each trial of an experiment
        session, along with all the iohub device events recorded by iohub
        during the session.

        Example psychopy code usage::

            # Load a trial handler and
            # create an associated table in the iohub data file
            #
            from psychopy.data import TrialHandler, importConditions

            exp_conditions=importConditions('trial_conditions.xlsx')
            trials = TrialHandler(exp_conditions, 1)

            # Inform the ioHub server about the TrialHandler
            #
            io.createTrialHandlerRecordTable(trials)

            # Read a row of the trial handler for
            # each trial of your experiment
            #
            for trial in trials:
                # do whatever...


            # During the trial, trial variable values can be updated
            #
            trial['TRIAL_START']=flip_time

            # At the end of each trial, before getting
            # the next trial handler row, send the trial
            # variable states to iohub so they can be stored for future
            # reference.
            #
            io.addTrialHandlerRecord(trial)

        """
        trial = trials.trialList[0]
        self._cv_order = cv_order
        if cv_order is None:
            self._cv_order = trial.keys()

        trial_condition_types = []

        for cond_name in self._cv_order:
            cond_val = trial[cond_name]
            if isinstance(cond_val, str):
                numpy_dtype = (cond_name, 'S', 256)
            elif isinstance(cond_val, int):
                numpy_dtype = (cond_name, 'i8')
            elif isinstance(cond_val, float):
                numpy_dtype = (cond_name, 'f8')
            else:
                numpy_dtype = (cond_name, 'S', 256)
            trial_condition_types.append(numpy_dtype)

        # pylint: disable=protected-access
        cvt_rpc = ('RPC', 'initConditionVariableTable',
                   (self.experimentID, self.experimentSessionID,
                    trial_condition_types))
        r = self._sendToHubServer(cvt_rpc)
        return r[2]

    def addTrialHandlerRecord(self, cv_row):
        """Adds the values from a TriaHandler row / record to the iohub data
        file for future data analysis use.

        :param cv_row:
        :return: None

        """
        data = []
        if isinstance(cv_row, (list, tuple)):
            data = list(cv_row)
        elif self._cv_order:
            for cv_name in self._cv_order:
                data.append(cv_row[cv_name])
        else:
            data = list(cv_row.values())

        for i, d in enumerate(data):
            if isinstance(d, str):
                data[i] = d.encode('utf-8')

        cvt_rpc = ('RPC', 'extendConditionVariableTable',
                   (self.experimentID, self.experimentSessionID, data))
        r = self._sendToHubServer(cvt_rpc)
        return r[2]

    def registerWindowHandles(self, *winHandles):
        """
        Sends 1 - n Window handles to iohub so it can determine if kb or
        mouse events were targeted at a psychopy window or other window.
        """
        r = self._sendToHubServer(('RPC', 'registerWindowHandles', winHandles))
        return r[2]

    def unregisterWindowHandles(self, *winHandles):
        """
        Sends 1 - n Window handles to iohub so it can determine if kb or
        mouse events were targeted at a psychopy window or other window.
        """
        r = self._sendToHubServer(
            ('RPC', 'unregisterWindowHandles', winHandles))
        return r[2]

    def updateWindowPos(self, win, x, y):
        r = self._sendToHubServer(('RPC', 'updateWindowPos', (win._hw_handle, (x, y))))
        return r[2]

    def getTime(self):
        """
        **Deprecated Method:** Use Computer.getTime instead. Remains here for
        testing time bases between processes only.
        """
        return self._sendToHubServer(('RPC', 'getTime'))[2]

    def setPriority(self, level='normal', disable_gc=False):
        """See Computer.setPriority documentation, where current process will
        be the iohub process."""
        return self._sendToHubServer(('RPC', 'setPriority',
                                      [level, disable_gc]))[2]

    def getPriority(self):
        """See Computer.getPriority documentation, where current process will
        be the iohub process."""
        return self._sendToHubServer(('RPC', 'getPriority'))[2]

    def getProcessAffinity(self):
        """
        Returns the current **ioHub Process** affinity setting,
        as a list of 'processor' id's (from 0 to getSystemProcessorCount()-1).
        A Process's Affinity determines which CPU's or CPU cores a process can
        run on. By default the ioHub Process can run on any CPU or CPU core.

        This method is not supported on OS X at this time.

        Args:
            None

        Returns:
            list: A list of integer values between 0 and
                  Computer.getSystemProcessorCount()-1, where values in the
                  list indicate processing unit indexes that the ioHub
                  process is able to run on.
        """
        r = self._sendToHubServer(('RPC', 'getProcessAffinity'))
        return r[2]

    def setProcessAffinity(self, processor_list):
        """
        Sets the **ioHub Process** Affinity based on the value of
        processor_list.

        A Process's Affinity determines which CPU's or CPU cores a process can
        run on. By default the ioHub Process can run on any CPU or CPU core.

        The processor_list argument must be a list of 'processor' id's;
        integers in the range of 0 to Computer.processing_unit_count-1,
        representing the processing unit indexes that the ioHub Server should
        be allowed to run on.

        If processor_list is given as an empty list, the ioHub Process will be
        able to run on any processing unit on the computer.

        This method is not supported on OS X at this time.

        Args:
            processor_list (list): A list of integer values between 0 and
                                   Computer.processing_unit_count-1,
                                   where values in the list indicate
                                   processing unit indexes that the ioHub
                                   process is able to run on.

        Returns:
            None
        """
        r = self._sendToHubServer(
            ('RPC', 'setProcessAffinity', processor_list))
        return r[2]

    def addDeviceToMonitor(self, device_class, device_config=None):
        """
        Normally this method should not be used, as all devices
        should be specified when the iohub server is being started.

        Adds a device to the ioHub Server for event monitoring during the
        experiment. Adding a device to the iohub server after it has been
        started can take 10'2 to 100's of msec to perform on the ioHub
        server (depending on the device type). When the device is being added,
        events from existing devices can not be monitored.

        Args:
            device_class (str): The iohub class name of the device being added.

            device_config (dict): The device configuration settings to be set.
                                  Device settings not provided in device_config
                                  will be set to the default values
                                  specified by the device.

        Returns:
            DeviceView Instance: The PsychoPy Process's view of the ioHub
                                 Device created that was created,
                                 as would be returned if a device was
                                 accessed using the .devices attribute
                                 or the .getDeviceByLabel() method.

        """
        if device_config is None:
            device_config = {}
        drpc = ('EXP_DEVICE', 'ADD_DEVICE', device_class, device_config)
        r = self._sendToHubServer(drpc)
        device_class_name, dev_name, _ = r[2]
        return self._addDeviceView(dev_name, device_class_name)

    def flushDataStoreFile(self):
        """Manually tell the iohub datastore to flush any events it has buffered in
        memory to disk. Any cached message events are sent to the iohub server
        before flushing the iohub datastore.

        Args:
            None

        Returns:
            None
        """
        self.sendMessageEvents()
        r = self._sendToHubServer(('RPC', 'flushIODataStoreFile'))
        return r

    def startCustomTasklet(self, task_name, task_class_path, **class_kwargs):
        """
        Instruct the iohub server to start running a custom tasklet given
        by task_class_path. It is important that the custom task does not block
        for any significant amount of time, or the processing of events by the
        iohub server will be negatively effected.

        See the customtask.py demo for an example of how to make a long running
        task not block the rest of the iohub server.
        """
        class_kwargs.setdefault('name', task_name)
        r = self._sendToHubServer(('CUSTOM_TASK', 'START', task_name,
                                   task_class_path, class_kwargs))
        return r

    def stopCustomTasklet(self, task_name):
        """
        Instruct the iohub server to stop the custom task that was previously
        started by calling self.startCustomTasklet(....). task_name identifies
        which custom task should be stopped and must match the task_name
        of a previously started custom task.
        """
        r = self._sendToHubServer(('CUSTOM_TASK', 'STOP', task_name))
        return r

    def shutdown(self):
        """Tells the ioHub Server to close all ioHub Devices, the ioDataStore,
        and the connection monitor between the PsychoPy and ioHub Processes.
        Then end the server process itself.

        Args:
            None

        Returns:
            None

        """
        self._shutDownServer()

    def quit(self):
        """Same as the shutdown() method, but has same name as PsychoPy
        core.quit() so maybe easier to remember."""
        self.shutdown()

    # Private Methods.....

    def _startServer(self, ioHubConfig=None, ioHubConfigAbsPath=None):
        """Starts the ioHub Process, storing it's process id, and creating the
        experiment side device representation for IPC access to public device
        methods."""
        experiment_info = None
        session_info = None
        hub_defaults_config = {}
        rootScriptPath = os.path.dirname(sys.argv[0])
        if len(rootScriptPath)<=1:
            rootScriptPath = os.path.abspath(".")
        # >>>>> Load / Create / Update iohub config file.....
        cfpath = os.path.join(IOHUB_DIRECTORY, 'default_config.yaml')
        with open(cfpath, 'r') as config_file:
            hub_defaults_config = yload(config_file, Loader=yLoader)

        if ioHubConfigAbsPath is None and ioHubConfig is None:
            ioHubConfig = dict(monitor_devices=[dict(Keyboard={}),
                                                dict(Display={}),
                                                dict(Mouse={})])
        elif ioHubConfig is not None and ioHubConfigAbsPath is None:
            if 'monitor_devices' not in ioHubConfig:
                raise KeyError("ioHubConfig must be provided with "
                               "'monitor_devices' key:value.")
            if 'data_store' in ioHubConfig:
                iods = ioHubConfig['data_store']
                if 'experiment_info' in iods and 'session_info' in iods:
                    experiment_info = iods['experiment_info']
                    session_info = iods['session_info']
                else:
                    raise KeyError("ERROR: ioHubConfig:ioDataStore must "
                                   "contain both a 'experiment_info' and a "
                                   "'session_info' entry.")
        elif ioHubConfigAbsPath is not None and ioHubConfig is None:
            with open(ioHubConfigAbsPath, 'r') as config_file:
                ioHubConfig = yload(config_file, Loader=yLoader)
        else:
            raise ValueError('Both a ioHubConfig dict object AND a path to an '
                             'ioHubConfig file can not be provided.')

        if ioHubConfig:
            updateDict(ioHubConfig, hub_defaults_config)

        if ioHubConfig and ioHubConfigAbsPath is None:
            if isinstance(ioHubConfig.get('monitor_devices'), dict):
                # short hand device spec is being used. Convert dict of
                # devices in a list of device dicts.
                devs = ioHubConfig.get('monitor_devices')
                devsList = [{dname: dc} for dname, dc in devs.items()]
                ioHubConfig['monitor_devices'] = devsList

            import tempfile
            with tempfile.NamedTemporaryFile(mode='w', suffix='iohub',
                                             delete=False) as tfile:
                tfile.write(json.dumps(ioHubConfig))
                ioHubConfigAbsPath = os.path.abspath(tfile.name)
        # <<<<< Finished Load / Create / Update iohub config file.

        self._iohub_server_config = ioHubConfig

        if sys.platform == 'darwin':
            self._osxKillAndFreePort()

        # >>>> Start iohub subprocess
        run_script = os.path.join(IOHUB_DIRECTORY, 'start_iohub_process.py')
        subprocessArgList = [sys.executable,
                             run_script,
                             '%.6f' % Computer.global_clock.getLastResetTime(),
                             rootScriptPath,
                             ioHubConfigAbsPath,
                             "{}".format(Computer.current_process.pid)]

        # To enable coverage in the iohub process, set the iohub\default_config
        # setting 'coverage_env_var' to the name of the coverage
        # config file that exists in the psychopy\iohub site-packages folder.
        # For example:
        #   coverage_env_var: .coveragerc
        #
        # If coverage_env_var is None or the file is not found,
        # coverage of ioHub Server process is disabled.
        coverage_env_var = self._iohub_server_config.get('coverage_env_var')

        envars = dict(os.environ)
        if coverage_env_var not in [None, 'None']:
            coverage_env_var = "{}".format(coverage_env_var)
            cov_config_path = os.path.join(IOHUB_DIRECTORY, coverage_env_var)
            if os.path.exists(cov_config_path):
                print("Coverage enabled for ioHub Server Process.")
            else:
                print("ioHub Process Coverage conf file not found: %s",
                      cov_config_path)
            envars['COVERAGE_PROCESS_START'] = coverage_env_var

        self._server_process = subprocess.Popen(subprocessArgList,
                                                env=envars,
                                                cwd=IOHUB_DIRECTORY,
                                                # set sub process stderr to be stdout so PsychoPy Runner
                                                # shows errors from iohub
                                                stderr=subprocess.STDOUT,
                                                )

        # Get iohub server pid and psutil process object
        # for affinity and process priority setting.
        Computer.iohub_process_id = self._server_process.pid
        Computer.iohub_process = psutil.Process(self._server_process.pid)

        # >>>>> Create open UDP port to ioHub Server
        server_udp_port = self._iohub_server_config.get('udp_port', 9000)
        from ..net import UDPClientConnection
        # initially open with a timeout so macOS does not hang.        
        self.udp_client = UDPClientConnection(remote_port=server_udp_port, timeout=0.1)

        # If ioHub server does not respond correctly,
        # terminate process and exit the program.
        if self._waitForServerInit() is False:
            self._server_process.terminate()
            return "ioHub startup failed."
        
        # close and reopen blocking version of socket
        self.udp_client.close()
        self.udp_client = UDPClientConnection(remote_port=server_udp_port)
        # <<<<< Done Creating open UDP port to ioHub Server

        # <<<<< Done starting iohub subprocess

        ioHubConnection.ACTIVE_CONNECTION = proxy(self)
        # Send iohub server any existing open psychopy window handles.
        try:
            from psychopy.visual import window
            window.IOHUB_ACTIVE = True
            if window.openWindows:
                whs = []
                # pylint: disable=protected-access
                for w in window.openWindows:
                    winfo = windowInfoDict(w())
                    whs.append(winfo)
                    w().backend.onMoveCallback = self.updateWindowPos
                self.registerWindowHandles(*whs)
        except ImportError:
            pass

        # Sending experiment_info if available.....
        if experiment_info:
            self._sendExperimentInfo(experiment_info)
        # Sending session_info if available.....
        if session_info:
            # print 'Sending session_info: {0}'.format(session_info)
            self._sendSessionInfo(session_info)

        # >>>> Creating client side iohub device wrappers...
        self._createDeviceList(ioHubConfig['monitor_devices'])

        return 'OK'

    def _waitForServerInit(self):
        # >>>> Wait for iohub server ready signal ....
        hubonline = False
        # timeout if ioServer does not reply in 30 seconds
        timeout_duration = self._iohub_server_config.get('start_process_timeout', 30.0)
        timeout_time = Computer.getTime() + timeout_duration
        while hubonline is False and Computer.getTime() < timeout_time:
            r = self._sendToHubServer(['GET_IOHUB_STATUS', ])
            if r:
                hubonline = r[1] == 'RUNNING'
            time.sleep(0.1)
        return hubonline

        # # <<<< Finished wait for iohub server ready signal ....

    def _createDeviceList(self, monitor_devices_config):
        """Create client side iohub device views.
        """
        # get the list of devices registered with the ioHub
        for device_config_dict in monitor_devices_config:
            device_class_name = list(device_config_dict.keys())[0]
            device_config = list(device_config_dict.values())[0]
            if device_config.get('enable', True) is True:
                try:
                    self._addDeviceView(device_class_name, device_config)
                except Exception: # pylint: disable=broad-except
                    print2err('_createDeviceList: Error adding class. ')
                    printExceptionDetailsToStdErr()

    def _addDeviceView(self, dev_cls_name, dev_config):
        """Add an iohub device view to self.devices"""
        try:
            name = dev_config.get('name', dev_cls_name.lower())
            dev_cls_name = "{}".format(dev_cls_name)
            dev_name = dev_cls_name.lower()
            cls_name_start = dev_name.rfind('.')
            dev_mod_pth = 'psychopy.iohub.devices.'
            if cls_name_start > 0:
                dev_mod_pth2 = dev_name[:cls_name_start]
                dev_mod_pth = '{0}{1}'.format(dev_mod_pth, dev_mod_pth2)
                dev_cls_name = dev_cls_name[cls_name_start + 1:]
            else:
                dev_mod_pth = '{0}{1}'.format(dev_mod_pth, dev_name)

            dev_import_result = import_device(dev_mod_pth, dev_cls_name)
            dev_cls, dev_cls_name, evt_cls_list = dev_import_result

            DeviceConstants.addClassMapping(dev_cls)

            device_event_ids = []
            for ev in list(evt_cls_list.values()):
                if ev.EVENT_TYPE_ID:
                    device_event_ids.append(ev.EVENT_TYPE_ID)
            EventConstants.addClassMappings(device_event_ids, evt_cls_list)

            name_start = name.rfind('.')
            if name_start > 0:
                name = name[name_start + 1:]

            from .. import client as iohubclientmod
            local_class = None
            local_module = getattr(iohubclientmod, dev_cls_name.lower(), False)
            if local_module:
                # need to touch local_module since it was lazy loaded

                # pylint: disable=exec-used
                exec('import psychopy.iohub.client.{}'.format(dev_cls_name.lower()))
                local_class = getattr(local_module, dev_cls_name, False)

            if local_class:
                d = local_class(self, dev_cls_name, dev_config)
            else:
                full_device_class_name = getFullClassName(dev_cls)[len('psychopy.iohub.devices.'):]
                full_device_class_name = full_device_class_name.replace('eyetracker.EyeTracker', 'EyeTracker')
                d = ioHubDeviceView(self, full_device_class_name, dev_cls_name, dev_config)

            self.devices.addDevice(name, d)
            return d
        except Exception: # pylint: disable=broad-except
            print2err('_addDeviceView: Error adding class. ')
            printExceptionDetailsToStdErr()
        return None

    def _convertDict(self, d):
        r = {}
        for k, v in d.items():
            if isinstance(v, bytes):
                v = str(v, 'utf-8')
            elif isinstance(v, list) or  isinstance(v, tuple):
                v = self._convertList(v)
            elif isinstance(v, dict):
                v = self._convertDict(v)
        
            if isinstance(k, bytes):
                k = str(k, 'utf-8')
            r[k]=v
        return r

    def _convertList(self, l):
        r = []
        for i in l:
            if isinstance(i, bytes):
                r.append(str(i, 'utf-8'))
            elif isinstance(i, list) or  isinstance(i, tuple):
                r.append(self._convertList(i))
            elif isinstance(i, dict):
                r.append(self._convertDict(i))
            else:
                r.append(i)
        return r

    def _sendToHubServer(self, tx_data):
        """General purpose local <-> iohub server process UDP based
        request - reply code. The method blocks until the request is fulfilled
        and and a response is received from the ioHub server.

        Args:
            tx_data (tuple): data to send to iohub server

        Return (object): response from the ioHub Server process.
        """
        try:
            # send request to host, return is # bytes sent.
            #print("SEND:",tx_data)
            self.udp_client.sendTo(tx_data)
        except Exception as e: # pylint: disable=broad-except
            import traceback
            traceback.print_exc()
            self.shutdown()
            raise e
        
        result = None
        
        try:
            # wait for response from ioHub server, which will be the
            # result data and iohub server address (ip4,port).
            result = self.udp_client.receive()
            if result:
                result, _ = result
            #print("RESULT:",result)
        except Exception as e: # pylint: disable=broad-except
            import traceback
            traceback.print_exc()
            self.shutdown()
            raise e

        # check if the reply is an error or not. If it is, raise the error.
        # TODO: This is not really working as planned, in part because iohub
        #       server does not consistently return error responses when needed
        errorReply = self._isErrorReply(result)
        if errorReply:           
            raise ioHubError(result)
        # Otherwise return the result
        
        if result is not None:
            # Use recursive conversion funcs                     
            if isinstance(result, list) or  isinstance(result, tuple):
                result = self._convertList(result)
            elif isinstance(result, dict):
                result = self._convertDict(result)
        return result

    def _sendExperimentInfo(self, experimentInfoDict):
        """Sends the experiment info from the experiment config file to the
        ioHub Server, which passes it to the ioDataStore, determines if the
        experiment already exists in the hdf5 file based on 'experiment_code',
        and returns a new or existing experiment ID based on that criteria.
        """
        fieldOrder = (('experiment_id', 0), ('code', ''), ('title', ''),
                      ('description', ''), ('version', ''))
        values = []
        for key, defaultValue in fieldOrder:
            if key in experimentInfoDict:
                values.append(experimentInfoDict[key])
            else:
                values.append(defaultValue)
                experimentInfoDict[key] = defaultValue

        r = self._sendToHubServer(('RPC', 'setExperimentInfo', (values,)))
        self.experimentID = r[2]
        experimentInfoDict['experiment_id'] = self.experimentID
        self._experimentMetaData = experimentInfoDict
        return r[2]

    def _sendSessionInfo(self, sess_info):
        """Sends the experiment session info from the experiment config file
        and the values entered into the session dialog to the ioHub Server,
        which passes it to the ioDataStore.

        The dataStore determines if the session already exists in the
        experiment file based on 'session_code', and returns a new
        session ID  if session_code is not in use by the experiment.
        """
        if self.experimentID is None:
            raise RuntimeError("Experiment ID must be set by calling"
                               " _sendExperimentInfo before calling"
                               " _sendSessionInfo.")
        if 'code' not in sess_info:
            raise ValueError("Code must be provided in sessionInfoDict"
                             " ( StringCol(24) ).")
        if 'name' not in sess_info:
            sess_info['name'] = ''
        if 'comments' not in sess_info:
            sess_info['comments'] = ''
        if 'user_variables' not in sess_info:
            sess_info['user_variables'] = {}

        org_sess_info = sess_info['user_variables']

        sess_info['user_variables'] = json.dumps(sess_info['user_variables'])
        r = self._sendToHubServer(('RPC', 'createExperimentSessionEntry',
                                   (sess_info,))
                                 )

        self.experimentSessionID = r[2]
        sess_info['user_variables'] = org_sess_info
        sess_info['session_id'] = self.experimentSessionID
        self._sessionMetaData = sess_info
        return sess_info['session_id']

    @staticmethod
    def eventListToObject(evt_data):
        """Convert an ioHub event currently in list value format into the
        correct ioHub.devices.DeviceEvent subclass for the given event type."""
        evt_type = evt_data[DeviceEvent.EVENT_TYPE_ID_INDEX]
        return EventConstants.getClass(evt_type).createEventAsClass(evt_data)

    @staticmethod
    def eventListToDict(evt_data):
        """Convert an ioHub event currently in list value format into
        the event as a dictionary of attribute name, attribute values."""
        if isinstance(evt_data, dict):
            return evt_data
        etype = evt_data[DeviceEvent.EVENT_TYPE_ID_INDEX]
        return EventConstants.getClass(etype).createEventAsDict(evt_data)


    @staticmethod
    def eventListToNamedTuple(evt_data):
        """Convert an ioHub event currently in list value format into the
         namedtuple format for an event."""
        if not isinstance(evt_data, list):
            return evt_data
        etype = evt_data[DeviceEvent.EVENT_TYPE_ID_INDEX]
        return EventConstants.getClass(etype).createEventAsNamedTuple(evt_data)

    # client utility methods.
    def _getDeviceList(self):
        r = self._sendToHubServer(('EXP_DEVICE', 'GET_DEVICE_LIST'))
        return r[2]

    def _shutDownServer(self):
        if self._shutdown_attempted is False:
            # send any cached experiment messages
            self.sendMessageEvents()

            try:
                from psychopy.visual import window
                window.IOHUB_ACTIVE = False
            except ImportError:
                pass

            self._shutdown_attempted = True
            TimeoutError = psutil.TimeoutExpired
            try:
                if self.udp_client:  # if it isn't already garbage-collected
                    self.udp_client.sendTo(('STOP_IOHUB_SERVER',))
                    self.udp_client.close()
                if Computer.iohub_process:
                    r = Computer.iohub_process.wait(timeout=5)
                    print('ioHub Server Process Completed With Code: ', r)
            except TimeoutError:
                print('Warning: TimeoutExpired, Killing ioHub Server process.')
                Computer.iohub_process.kill()
            except Exception:  # pylint: disable=broad-except
                print("Warning: Unhandled Exception. "
                      "Killing ioHub Server process.")
                if Computer.iohub_process:
                    Computer.iohub_process.kill()
                printExceptionDetailsToStdErr()
            finally:
                ioHubConnection.ACTIVE_CONNECTION = None
                self._server_process = None
                Computer.iohub_process_id = None
                Computer.iohub_process = None
            return True

    @staticmethod
    def _isErrorReply(data):
        """
        Check if an iohub server reply contains an error that should be raised
        by the local process.
        """
        if isIterable(data) and len(data) > 0:
            d0 = data[0]
            if isIterable(d0):
                return False
            else:
                if isinstance(d0, str) and d0.find('ERROR') >= 0:
                    return data
                return False
        else:
            return data #'Invalid Response Received from ioHub Server'


    def _osxKillAndFreePort(self):
        server_udp_port = self._iohub_server_config.get('udp_port', 9000)
        p = subprocess.Popen(['lsof', '-i:%d'%server_udp_port, '-P'],
                             stdout=subprocess.PIPE)
        lines = p.communicate()[0]
        for line in lines.splitlines():
            if line.startswith('Python'):
                PID, userID = line.split()[1:3]
                # could verify same userID as current user, probably not needed
                os.kill(int(PID), signal.SIGKILL)
                print('Called  os.kill(int(PID),signal.SIGKILL): ', PID, userID)

    def __del__(self):
        try:
            self._shutDownServer()
            ioHubConnection.ACTIVE_CONNECTION = None
        except Exception: # pylint: disable=broad-except
            pass

##############################################################################

class ioEvent():
    """
    Parent class for all events generated by a psychopy.iohub.client
    Device wrapper.
    """
    _attrib_index = dict()
    _attrib_index['id'] = DeviceEvent.EVENT_ID_INDEX
    _attrib_index['time'] = DeviceEvent.EVENT_HUB_TIME_INDEX
    _attrib_index['type'] = DeviceEvent.EVENT_TYPE_ID_INDEX

    def __init__(self, ioe_array, device=None):
        self._time = ioe_array[ioEvent._attrib_index['time']]
        self._id = ioe_array[ioEvent._attrib_index['id']]
        self._type = ioe_array[ioEvent._attrib_index['type']]
        self._device = device

    @property
    def device(self):
        """
        The ioHubDeviceView that is associated with the event, i.e. the
        iohub device view for the device that generated the event.

        :return: ioHubDeviceView

        """
        return self._device

    @property
    def time(self):
        """
        The time stamp of the event. Uses the same time base that is used by
        psychopy.core.getTime()

        :return: float
        """
        return self._time

    @property
    def id(self):
        """The unique id for the event; in some cases used to track associated
        events.

        :return: int

        """
        return self._id

    @property
    def type(self):
        """The event type string constant.

        :return: str

        """
        return EventConstants.getName(self._type)

    @property
    def dict(self):
        d = {}
        for k in self._attrib_index:
            d[k] = getattr(self, k)
        return d

    def __str__(self):
        return 'time: %.3f, type: %s, id: %d' % (self.time,
                                                 self.type,
                                                 self.id)

_lazyImports = """
from psychopy.iohub.client.connect import launchHubServer
from psychopy.iohub.client import keyboard
from psychopy.iohub.client import wintab
"""

try:
    lazy_import(globals(), _lazyImports)
except Exception as e: #pylint: disable=broad-except
    print2err('lazy_import Exception:', e)
    exec(_lazyImports) #pylint: disable=exec-used



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from psychopy.visual import Window, ShapeStim
from psychopy import event, core, monitors
from psychopy.constants import NOT_STARTED
import pyglet
from pyglet.window.mouse import LEFT, MIDDLE, RIGHT
try:
    import pygame
    havePygame = True
except Exception:
    havePygame = False
import pytest
import copy
import threading
import numpy as np
from psychopy.tests import skip_under_vm

"""test with both pyglet and pygame:
    cd psychopy/psychopy/
    py.test -k event --cov-report term-missing --cov event.py
"""


class DelayedFakeKeys(threading.Thread):
    def __init__(self, keys, modifiers=0, delay=.01):
        threading.Thread.__init__(self, None, 'fake key', None)
        if isinstance(keys, str):
            self.keys = [keys]
        else:
            self.keys = keys
        self.modifiers = modifiers
        self.delay = delay

    def run(self):
        core.wait(self.delay)
        [event._onPygletKey(key, modifiers=self.modifiers, emulated=True)
         for key in self.keys]


class DelayedAddFakeKeysToBuffer(threading.Thread):
    def __init__(self, keys, modifiers=0, delay=.01):
        threading.Thread.__init__(self, None, 'fake key', None)
        if isinstance(keys, str):
            self.keys = [keys]
        else:
            self.keys = keys
        self.modifiers = modifiers
        self.delay = delay

    def run(self):
        core.wait(self.delay)
        fake_events = [(key, self.modifiers, -1) for key in self.keys]
        event._keyBuffer.extend(fake_events)

class _baseTest():
    #this class allows others to be created that inherit all the tests for
    #a different window config
    @classmethod
    def setup_class(self):#run once for each test class (window)
        self.win=None
        self.contextName
        raise NotImplementedError
    @classmethod
    def teardown_class(self):#run once for each test class (window)
        try:
            self.win.close()
        except AttributeError:
            pass

    def test_mouse_pos(self):
        if self.win.winType == 'pygame':
            pytest.skip()  # pygame.setVisible errors
        for w in (self.win,): #, None):
            for p in (None, (0,0)):
                m = event.Mouse(newPos=p, win=w)
                assert m.units == 'norm'
                m.setPos((0,0))
                m.getPos()

    def test_emulated_mouse(self):
        mouse = event.Mouse()  # real mouse
        event.mouseButtons = [0,0,0]
        [c.reset() for c in event.mouseClick]  # mouse click RT clocks
        assert not any(event.mouseButtons)
        assert not any(event.mouseTimes)

        # fake clicks on all buttons:
        event._onPygletMousePress(0, 0, LEFT | MIDDLE | RIGHT, None, emulated=True)
        assert all(mouse.getPressed())
        assert all([RT < 0.01 for RT in event.mouseTimes])  # should be < .0001

        # fake release all buttons:
        event._onPygletMouseRelease(0, 0, LEFT | MIDDLE | RIGHT, None, emulated=True)
        assert not any(event.mouseButtons)

    def test_mouse_clock(self):
        x, y = 0, 0
        scroll_x, scroll_y = 1, 1
        dx, dy = 1, 1
        zeros = [0, 0, 0]
        for b in [pyglet.window.mouse.LEFT, pyglet.window.mouse.MIDDLE, pyglet.window.mouse.RIGHT]:
            event.mouseButtons = copy.copy(zeros)
            event.mouseTimes = copy.copy(zeros)
            event._onPygletMousePress(x,y, b, None)
            assert event.mouseButtons != zeros
            assert event.mouseTimes != zeros
            event._onPygletMouseRelease(x,y, b, None)
            assert event.mouseButtons == zeros
        event._onPygletMouseWheel(x,y,scroll_x, scroll_y)
        event._onPygletMouseMotion(x, y, dx, dy)
        event.startMoveClock()
        event.stopMoveClock()
        event.resetMoveClock()

        m = event.Mouse()
        assert m.mouseMoveTime() >= 0
        t = 0.05
        core.wait(t)
        assert t - 0.01 < m.mouseMoveTime() < t + 0.01

    def test_clearEvents(self):
        for t in ['mouse', 'joystick', 'keyboard', None]:
            event.clearEvents(t)

    def test_clearEvents_keyboard(self):
        event._onPygletKey(symbol='x', modifiers=0, emulated=True)
        event.clearEvents('keyboard')
        assert not event._keyBuffer

    def test_clearEvents_mouse(self):
        """Keyboard buffer should not be affected.
        """
        event._onPygletKey(symbol='x', modifiers=0, emulated=True)
        event.clearEvents('mouse')
        assert event._keyBuffer

    def test_clearEvents_joystick(self):
        """Keyboard buffer should not be affected.
        """
        event._onPygletKey(symbol='x', modifiers=0, emulated=True)
        event.clearEvents('joystick')
        assert event._keyBuffer

    @skip_under_vm
    def test_keys(self):
        if self.win.winType == 'pygame':
            pytest.skip()
        event.clearEvents()
        assert event.getKeys() == []
        for k in ['s', 'return']:
            event.clearEvents()
            event._onPygletKey(symbol=k, modifiers=0, emulated=True)
            assert k in event.getKeys()
            event._onPygletKey(symbol=17, modifiers=0, emulated=False)
            assert '17' in event.getKeys()

            # test that key-based RT is about right
            event.clearEvents()
            c = core.Clock()
            t = 0.05
            core.wait(t)
            event._onPygletKey(symbol=k, modifiers=0, emulated=True)
            resp = event.getKeys(timeStamped=c)
            assert k in resp[0][0]
            assert t - 0.01 < resp[0][1] < t + 0.01

            event._onPygletKey(symbol=k, modifiers=0, emulated=True)
            assert k in event.getKeys(timeStamped=True)[0]
            event._onPygletKey(symbol=k, modifiers=0, emulated=True)
            event._onPygletKey(symbol='x', modifiers=0, emulated=True)  # nontarget
            assert k in event.getKeys(keyList=[k, 'd'])

            # waitKeys implicitly clears events, so use a thread to add a delayed key press
            assert event.waitKeys(maxWait=-1) is None
            keyThread = DelayedFakeKeys(k)
            keyThread.start()
            assert event.waitKeys(maxWait=.1) == [k]
            keyThread = DelayedFakeKeys(k)
            keyThread.start()
            assert event.waitKeys(maxWait=.1, keyList=[k]) == [k]

            # test time-stamped waitKeys
            c = core.Clock()
            delay=0.01
            keyThread = DelayedFakeKeys(k, delay=delay)
            keyThread.start()
            result = event.waitKeys(maxWait=.1, keyList=[k], timeStamped=c)
            assert result[0][0] == k
            assert result[0][1] - delay < .01  # should be ~0 except for execution time

    @skip_under_vm
    def test_waitKeys_clearEvents_True(self):
        key = 'x'
        DelayedAddFakeKeysToBuffer(key).start()
        key_events = event.waitKeys(clearEvents=True)
        assert key_events == [key]

    @skip_under_vm
    def test_waitKeys_clearEvents_False(self):
        keys = ['x', 'y', 'z']
        [event._onPygletKey(symbol=key, modifiers=0, emulated=True)
         for key in keys]

        key_events = event.waitKeys(keyList=keys[1:], clearEvents=False)
        assert 'x' not in key_events
        assert 'y' in key_events
        assert 'z' in key_events

    @skip_under_vm
    def test_waitKeys_keyList_clearEvents_True(self):
        keys = ['x', 'y', 'z']
        DelayedAddFakeKeysToBuffer(keys).start()
        key_events = event.waitKeys(keyList=keys[:-1], clearEvents=True)

        assert 'x' in key_events
        assert 'y' in key_events
        assert 'z' not in key_events
        assert 'z' in event.getKeys()

    def test_xydist(self):
        assert event.xydist([0,0], [1,1]) == np.sqrt(2)

    @skip_under_vm
    def test_mouseMoved(self):
        m = event.Mouse()
        m.prevPos = [0, 0]
        m.lastPos = [0, 1]
        assert m.mouseMoved()  # call to mouseMoved resets prev and last

        m.prevPos = [0, 0]
        m.lastPos = [0, 1]
        assert m.mouseMoved(distance=0.5)
        for reset in [True, 'here', (1,2)]:
            assert not m.mouseMoved(reset=reset)

    def test_set_visible(self):
        if self.win.winType == 'pygame':
            pytest.skip()
        m = event.Mouse()
        for v in (0, 1):
            m.setVisible(v)
            w = self.win
            m.win = None
            m.setVisible(v)
            m.win = w

    def test_misc(self):
        m = event.Mouse()
        m.getRel()
        m.getWheelRel()
        m.getVisible()
        m.clickReset()
        # to-do: proper test of mouseClick and mouseTimes being changed

        # not much to assert here:
        m.getPressed()
        m.getPressed(getTime=True)

    def test_isPressedIn(self):
        m = event.Mouse(self.win, newPos=(0,0))
        s = ShapeStim(self.win, vertices=[[10,10],[10,-10],[-10,-10],[-10,10]], autoLog=False)
        if not s.contains(m.getPos()):
            pytest.skip()  # or can't test

        event.mouseButtons = [1, 1, 1]
        assert m.isPressedIn(s)

    # obsolete?
    # m._pix2windowUnits()
    # m._windowUnits2pix()

    def test_builder_key_resp(self):
        # just inits
        bk = event.BuilderKeyResponse()
        assert bk.status == NOT_STARTED
        assert bk.keys == [] #the key(s) pressed
        assert bk.corr == 0 #was the resp correct this trial? (0=no, 1=yes)
        assert bk.rt == [] #response time(s)
        assert bk.clock.getTime() < .001


@pytest.mark.event
class TestPygletNorm(_baseTest):
    @classmethod
    def setup_class(self):
        mon = monitors.Monitor('testMonitor')
        mon.setDistance(10.0) #exaggerate the effect of flatness by setting the monitor close
        mon.setWidth(40.0)
        mon.setSizePix([1024,768])
        self.win = Window([128,128], monitor=mon, winType='pyglet', pos=[50,50], autoLog=False)
        if havePygame:
            assert pygame.display.get_init() == 0


class xxxTestPygameNorm(_baseTest):
    @classmethod
    def setup_class(self):
        self.win = Window([128,128], winType='pygame', pos=[50,50], autoLog=False)
        assert pygame.display.get_init() == 1
        assert event.havePygame


if __name__ == '__main__':
    import pytest
    pytest.main()



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

try:
    import pyHook
except ImportError:
    import pyWinhook as pyHook
import win32api
import ctypes
from unicodedata import category as ucategory
from . import ioHubKeyboardDevice, psychopy_key_mappings
from ...constants import KeyboardConstants, EventConstants
from .. import Computer, Device
from ...errors import print2err, printExceptionDetailsToStdErr

win32_vk = KeyboardConstants._virtualKeyCodes

getTime = Computer.getTime

jdumps = lambda x: str(x)
try:
    import ujson
    jdumps = ujson.dumps
except Exception:
    import json
    jdumps = json.dumps

# Map key value when numlock is ON
# to value when numlock is OFF.
numpad_key_value_mappings = dict(Numpad0='insert',
                                 Numpad1='end',
                                 Numpad2='down',
                                 Numpad3='pagedown',
                                 Numpad4='left',
                                 Numpad5=' ',
                                 Numpad6='right',
                                 Numpad7='home',
                                 Numpad8='up',
                                 Numpad9='pageup',
                                 Decimal='delete'
                                 )


def updateToPsychopyKeymap():
    global numpad_key_value_mappings
    numpad_key_value_mappings = dict(Numpad0='num_0',
                                     Numpad1='num_1',
                                     Numpad2='num_2',
                                     Numpad3='num_3',
                                     Numpad4='num_4',
                                     Numpad5='num_5',
                                     Numpad6='num_6',
                                     Numpad7='num_7',
                                     Numpad8='num_8',
                                     Numpad9='num_9',
                                     Decimal='num_decimal'
                                     )

class Keyboard(ioHubKeyboardDevice):
    _win32_modifier_mapping = {
        win32_vk.VK_LCONTROL: 'lctrl',
        win32_vk.VK_RCONTROL: 'rctrl',
        win32_vk.VK_LSHIFT: 'lshift',
        win32_vk.VK_RSHIFT: 'rshift',
        win32_vk.VK_LMENU: 'lalt',
        win32_vk.VK_RMENU: 'ralt',
        win32_vk.VK_LWIN: 'lcmd',
        win32_vk.VK_RWIN: 'rcmd',
        win32_vk.VK_CAPITAL: 'capslock',
        win32_vk.VK_NUM_LOCK: 'numlock',
        win32_vk.VK_SCROLL: 'scrolllock'
    }

    __slots__ = ['_user32', '_keyboard_state', '_unichar']

    def __init__(self, *args, **kwargs):
        ioHubKeyboardDevice.__init__(self, *args, **kwargs['dconfig'])
        self._user32 = ctypes.windll.user32
        self._keyboard_state = (ctypes.c_ubyte * 256)()
        self._unichar = (ctypes.c_wchar * 8)()

        if self.use_psychopy_keymap:
            updateToPsychopyKeymap()

        self.resetKeyAndModState()

    def resetKeyAndModState(self):
        for i in range(256):
            self._keyboard_state[i] = 0

        ioHubKeyboardDevice._modifier_value = 0
        for stateKeyID in [
                win32_vk.VK_SCROLL,
                win32_vk.VK_NUM_LOCK,
                win32_vk.VK_CAPITAL]:
            state = pyHook.GetKeyState(stateKeyID)
            if state:
                self._keyboard_state[stateKeyID] = state
                modKeyName = Keyboard._win32_modifier_mapping.get(
                    stateKeyID, None)
                mod_value = KeyboardConstants._modifierCodes.getID(modKeyName)
                ioHubKeyboardDevice._modifier_value += mod_value

    def _updateKeyMapState(self, event):
        keyID = event.KeyID
        is_press = event.Type == EventConstants.KEYBOARD_PRESS

        if is_press:
            self._keyboard_state[keyID] = 0x80
        else:
            self._keyboard_state[keyID] = 0

        if event.cap_state and (
            self._keyboard_state[
                win32_vk.VK_CAPITAL] & 1) == 0:
            self._keyboard_state[win32_vk.VK_CAPITAL] += 1
        elif not event.cap_state and (self._keyboard_state[win32_vk.VK_CAPITAL] & 1) == 1:
            self._keyboard_state[win32_vk.VK_CAPITAL] -= 1

        if event.scroll_state and (
            self._keyboard_state[
                win32_vk.VK_SCROLL] & 1) == 0:
            self._keyboard_state[win32_vk.VK_SCROLL] += 1
        elif not event.scroll_state and (self._keyboard_state[win32_vk.VK_SCROLL] & 1) == 1:
            self._keyboard_state[win32_vk.VK_SCROLL] -= 1

        if event.num_state and (
            self._keyboard_state[
                win32_vk.VK_NUM_LOCK] & 1) == 0:
            self._keyboard_state[win32_vk.VK_NUM_LOCK] += 1
        elif not event.num_state and (self._keyboard_state[win32_vk.VK_NUM_LOCK] & 1) == 1:
            self._keyboard_state[win32_vk.VK_NUM_LOCK] -= 1

        modKeyName = Keyboard._win32_modifier_mapping.get(keyID, None)
        if modKeyName:
            if is_press:
                if keyID in [win32_vk.VK_LSHIFT, win32_vk.VK_RSHIFT]:
                    self._keyboard_state[win32_vk.VK_SHIFT] = 0x80
                elif keyID in [win32_vk.VK_LCONTROL, win32_vk.VK_RCONTROL]:
                    self._keyboard_state[win32_vk.VK_CONTROL] = 0x80
                elif keyID in [win32_vk.VK_LMENU, win32_vk.VK_RMENU]:
                    self._keyboard_state[win32_vk.VK_MENU] = 0x80
            else:
                if keyID in [win32_vk.VK_LSHIFT, win32_vk.VK_RSHIFT]:
                    self._keyboard_state[win32_vk.VK_SHIFT] = 0
                elif keyID in [win32_vk.VK_LCONTROL, win32_vk.VK_RCONTROL]:
                    self._keyboard_state[win32_vk.VK_CONTROL] = 0
                elif keyID in [win32_vk.VK_LMENU, win32_vk.VK_RMENU]:
                    self._keyboard_state[win32_vk.VK_MENU] = 0

        return modKeyName

    def _updateModValue(self, keyID, is_press):
        modKeyName = Keyboard._win32_modifier_mapping.get(keyID, None)
        if modKeyName:
            mod_value = KeyboardConstants._modifierCodes.getID(modKeyName)
            mod_set = ioHubKeyboardDevice._modifier_value&mod_value == mod_value
 
            if keyID not in [win32_vk.VK_CAPITAL, win32_vk.VK_SCROLL,
                             win32_vk.VK_NUM_LOCK]:
                if is_press and not mod_set:
                    ioHubKeyboardDevice._modifier_value += mod_value
                elif not is_press and mod_set:
                    ioHubKeyboardDevice._modifier_value -= mod_value
            else:
                if is_press:
                    if mod_set:
                        ioHubKeyboardDevice._modifier_value -= mod_value
                    else:
                        ioHubKeyboardDevice._modifier_value += mod_value
        return ioHubKeyboardDevice._modifier_value

    def _getKeyCharValue(self, event):
        key = None
        char = ''
        ucat = ''

        # Get char value
        #
        result = self._user32.ToUnicode(event.KeyID, event.ScanCode,
                                        ctypes.byref(self._keyboard_state),
                                        ctypes.byref(self._unichar), 8, 0)

        if result > 0:
            char = self._unichar[result - 1].encode('utf-8')
            ucat = ucategory(self._unichar[result - 1])

        # Get .key value
        #
        if event.Key in numpad_key_value_mappings:
            key = numpad_key_value_mappings[event.Key]
        elif ucat.lower() != 'cc':
            prev_shift = self._keyboard_state[win32_vk.VK_SHIFT]
            prev_numlock = self._keyboard_state[win32_vk.VK_NUM_LOCK]
            prev_caps = self._keyboard_state[win32_vk.VK_CAPITAL]
            self._keyboard_state[win32_vk.VK_SHIFT] = 0
            self._keyboard_state[win32_vk.VK_NUM_LOCK] = 0
            result = self._user32.ToUnicode(event.KeyID, event.ScanCode,
                                            ctypes.byref(self._keyboard_state),
                                            ctypes.byref(self._unichar), 8, 0)
            self._keyboard_state[win32_vk.VK_SHIFT] = prev_shift
            self._keyboard_state[win32_vk.VK_NUM_LOCK] = prev_numlock
            self._keyboard_state[win32_vk.VK_CAPITAL] = prev_caps
            if result > 0:
                key = self._unichar[result - 1].encode('utf-8')

        if key is None:
            key = KeyboardConstants._getKeyName(event)

        if isinstance(key, bytes):
            key = str(key, 'utf-8')
        if isinstance(char, bytes):
            char = str(char, 'utf-8')

        key = key.lower()

        # misc. char value cleanup.
        if key == 'return':
            char = '\n'.encode('utf-8')
        elif key in ('escape', 'backspace'):
            char = ''

        if Keyboard.use_psychopy_keymap and key in psychopy_key_mappings.keys():
            key = psychopy_key_mappings[key]

            # win32 specific handling of keypad / and - keys
            if event.Key == 'Subtract':
                key = 'num_subtract'
            elif event.Key == 'Divide':
                key = 'num_divide'

        return key, char

    def _evt2json(self, event):
        return jdumps(dict(Type=event.Type,
                           Time=event.Time,
                           KeyID=event.KeyID,
                           ScanCode=event.ScanCode,
                           Ascii=event.Ascii,
                           flags=event.flags,
                           Key=event.Key,
                           scroll_state=event.scroll_state,
                           num_state=event.num_state,
                           cap_state=event.cap_state))

    def _addEventToTestLog(self, event_data):
        if self._log_events_file is None:
            import datetime
            cdate = datetime.datetime.now().strftime('%Y_%m_%d_%H_%M')
            self._log_events_file = open(
                'win32_events_{0}.log'.format(cdate), 'w')
        self._log_events_file.write(self._evt2json(event_data) + '\n')

    def _nativeEventCallback(self, event):
        if self.isReportingEvents():
            notifiedTime = getTime()

            report_system_wide_events = self.getConfiguration().get(
                'report_system_wide_events', True)
            if report_system_wide_events is False:
                pyglet_window_hnds = self._iohub_server._psychopy_windows.keys()
                if len(
                        pyglet_window_hnds) > 0 and event.Window not in pyglet_window_hnds:
                    return True

            event.Type = EventConstants.KEYBOARD_RELEASE
            if event.Message in [pyHook.HookConstants.WM_KEYDOWN,
                                 pyHook.HookConstants.WM_SYSKEYDOWN]:
                event.Type = EventConstants.KEYBOARD_PRESS

            self._last_callback_time = notifiedTime

            event.RepeatCount = 0
            key_already_pressed = self._key_states.get(event.KeyID, None)
            if key_already_pressed and event.Type == EventConstants.KEYBOARD_PRESS:
                event.RepeatCount = key_already_pressed[1] + 1
                if self._report_auto_repeats is False and event.RepeatCount > 0:
                    return True

            event.Modifiers = 0
            event.scroll_state = pyHook.GetKeyState(win32_vk.VK_SCROLL)
            event.num_state = pyHook.GetKeyState(win32_vk.VK_NUM_LOCK)
            event.cap_state = pyHook.GetKeyState(win32_vk.VK_CAPITAL)

            if self.getConfiguration().get('log_events_for_testing', False):
                self._addEventToTestLog(event)

            self._addNativeEventToBuffer((notifiedTime, event))
        # pyHook require the callback to return True to inform the windows
        # low level hook functionality to pass the event on.
        return True

    def _getIOHubEventObject(self, native_event_data):
        try:
            notifiedTime, event = native_event_data
            is_press = event.Type == EventConstants.KEYBOARD_PRESS
            keyID = event.KeyID
            device_time = event.Time / 1000.0  # convert to sec
            time = notifiedTime
            # since this is a keyboard device using a callback method,
            # confidence_interval is not applicable
            confidence_interval = 0.0
            # since this is a keyboard, we 'know' there is a delay, but until
            # we support setting a delay in the device properties based on
            # external testing for a given keyboard, we will leave at 0.
            delay = 0.0
            modKeyName = self._updateKeyMapState(event)
            event.Modifiers = self._updateModValue(keyID, is_press)

            # Get key and char fields.....

            if modKeyName:
                key = modKeyName
                char = ''
            else:
                key, char = self._getKeyCharValue(event)

            kb_event = [0,
                        0,
                        0,  # device id (not currently used)
                        Device._getNextEventID(),
                        event.Type,
                        device_time,
                        notifiedTime,
                        time,
                        confidence_interval,
                        delay,
                        0,
                        event.RepeatCount,
                        event.ScanCode,
                        event.KeyID,
                        0,
                        key,
                        event.Modifiers,
                        event.Window,
                        char,  # .char
                        0.0,  # duration
                        0  # press_event_id
                        ]

            ioHubKeyboardDevice._updateKeyboardEventState(self, kb_event,
                                                          is_press)
            return kb_event
        except Exception:
            printExceptionDetailsToStdErr()

    def _syncPressedKeyState(self):
        remove_keys = []
        for kid in self._key_states.keys():
            if win32api.GetAsyncKeyState(kid) == 0:
                # Key is no longer pressed, remove it from pressed key dict
                remove_keys.append(kid)
        for kid in remove_keys:
            del self._key_states[kid]



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from . import ioHubKeyboardDevice, psychopy_key_mappings
from .. import Computer, Device
from ...constants import EventConstants
from ...errors import printExceptionDetailsToStdErr, print2err

getTime = Computer.getTime

NUMLOCK_MODIFIER = 8192

psychopy_numlock_key_mappings = dict()
psychopy_numlock_key_mappings['num_end'] = 'num_1'
psychopy_numlock_key_mappings['num_down'] = 'num_2'
psychopy_numlock_key_mappings['num_page_down'] = 'num_3'
psychopy_numlock_key_mappings['num_left'] = 'num_4'
psychopy_numlock_key_mappings['num_begin'] = 'num_5'
psychopy_numlock_key_mappings['num_right'] = 'num_6'
psychopy_numlock_key_mappings['num_home'] = 'num_7'
psychopy_numlock_key_mappings['num_up'] = 'num_8'
psychopy_numlock_key_mappings['num_page_up'] = 'num_9'
psychopy_numlock_key_mappings['num_insert'] = 'num_0'
psychopy_numlock_key_mappings['num_delete'] = 'num_decimal'

class Keyboard(ioHubKeyboardDevice):
    event_id_index = None

    def __init__(self, *args, **kwargs):
        ioHubKeyboardDevice.__init__(self, *args, **kwargs['dconfig'])

        if self.event_id_index is None:
            from . import KeyboardInputEvent
            Keyboard.auto_repeated_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'auto_repeated')
            Keyboard.key_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'key')
            Keyboard.key_id_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'key_id')
            Keyboard.event_type_index = KeyboardInputEvent.EVENT_TYPE_ID_INDEX
            Keyboard.event_modifiers_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'modifiers')
            Keyboard.win_id_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index(
                'window_id')
            Keyboard.event_id_index = KeyboardInputEvent.EVENT_ID_INDEX

    def _nativeEventCallback(self, event):
        try:
            self._last_callback_time = getTime()
            if self.isReportingEvents():
                event_array = event[0]
                
                key = event_array[Keyboard.key_index]
                if isinstance(key, bytes):
                    event_array[Keyboard.key_index] = key = str(key, 'utf-8')                
                
                if Keyboard.use_psychopy_keymap:
                    if key in psychopy_key_mappings.keys():
                        key = event_array[Keyboard.key_index] = psychopy_key_mappings[key]
                    elif key == 'num_next':
                        key = event_array[Keyboard.key_index] = 'num_page_down'
                    elif key == 'num_prior':
                        key = event_array[Keyboard.key_index] = 'num_page_up'

                    if (event_array[Keyboard.event_modifiers_index]&NUMLOCK_MODIFIER) > 0:
                        if key in psychopy_numlock_key_mappings:
                            key = event_array[Keyboard.key_index] = psychopy_numlock_key_mappings[key]  
                            
                # Check if key event window id is in list of psychopy
                # windows and what report_system_wide_events value is
                report_system_wide_events = self.getConfiguration().get(
                    'report_system_wide_events', True)
                if report_system_wide_events is False:
                    pyglet_window_hnds = self._iohub_server._psychopy_windows.keys()
                    if len(pyglet_window_hnds) > 0 and event_array[self.win_id_index] not in pyglet_window_hnds:
                        return True

                is_pressed = event_array[
                    self.event_type_index] == EventConstants.KEYBOARD_PRESS

                if is_pressed and self._report_auto_repeats is False:
                    # AUto repeat value provided by pyXHook code
                    auto_repeat_count = event_array[self.auto_repeated_index]
                    if auto_repeat_count > 0:
                        return True

                # set event id for event since it has passed all filters
                event_array[self.event_id_index] = Device._getNextEventID()
                ioHubKeyboardDevice._modifier_value = event_array[
                    self.event_modifiers_index]

                self._updateKeyboardEventState(event_array, is_pressed)

                self._addNativeEventToBuffer(event_array)
        except Exception:
            printExceptionDetailsToStdErr()
        return 1

    def _getIOHubEventObject(self, native_event_data):
        return native_event_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import sys
from .errors import print2err, printExceptionDetailsToStdErr
from .util import module_directory

if sys.platform == 'darwin':
    import objc  # pylint: disable=import-error

EXP_SCRIPT_DIRECTORY = ''


def _localFunc():
    return None


IOHUB_DIRECTORY = module_directory(_localFunc)

_DATA_STORE_AVAILABLE = False
try:
    import tables
    _DATA_STORE_AVAILABLE = True
except ImportError:
    print2err('WARNING: pytables package not found. ',
              'ioHub functionality will be disabled.')
except Exception:
    printExceptionDetailsToStdErr()

from psychopy.iohub.constants import EventConstants, KeyboardConstants, MouseConstants

lazyImports = """
from psychopy.iohub.client.connect import launchHubServer
from psychopy.iohub.devices.computer import Computer
from psychopy.iohub.client.eyetracker.validation import ValidationProcedure
"""

try:
    from psychopy.contrib.lazy_import import lazy_import
    lazy_import(globals(), lazyImports)
except Exception:
    exec(lazyImports)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of several visual stimuli depending on the mouse position.
"""

from psychopy import visual, event, core
import numpy

win = visual.Window([600, 600], color='black')

gabor = visual.GratingStim(win, mask='gauss', pos=[-0.5, -0.5], color=[0, 0, 1], sf=5, ori=30)
movie = visual.MovieStim3(win, 'jwpIntro.mp4', units='pix', pos=[100, 100], size=[160, 120])
txt = u"unicode (eg \u03A8 \u040A \u03A3)"
text = visual.TextStim(win, pos=[0.5, -0.5], text=txt, font=['Times New Roman'])
faceRGB = visual.ImageStim(win, image='face.jpg', pos=[-0.5, 0.5])

mouse = event.Mouse()
instr = visual.TextStim(win, text='move the mouse around')

t = 0.0
while not event.getKeys() and not mouse.getPressed()[0]:
    # get mouse events
    mouse_dX, mouse_dY = mouse.getRel()

    gabor.ori -= mouse_dY * 10
    text.ori += mouse_dY * 10
    faceRGB.ori += mouse_dY * 10
    movie.ori -= mouse_dY * 10

    t += 1/60.0
    gabor.phase = t * 2.0
    gabor.draw()
    text.color = [numpy.sin(t * 2), 0, 1]
    text.draw()
    faceRGB.draw()
    movie.draw()

    instr.draw()

    win.flip()

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Extensible set of components for the PsychoPy Builder view
"""

import os
import glob
import copy
import shutil
from os.path import join, dirname, abspath, split
from importlib import import_module  # helps python 2.7 -> 3.x migration
from ._base import BaseVisualComponent, BaseComponent
from ..params import Param
from psychopy.localization import _translate
from psychopy.experiment import py2js
import psychopy.logging as logging

excludeComponents = ['BaseComponent', 'BaseVisualComponent', 'BaseStandaloneRoutine'  # templates only
                     ]  # this one isn't ready yet

pluginComponents = {}  # components registered by loaded plugins

# try to remove old pyc files in case they're detected as components
pycFiles = glob.glob(join(split(__file__)[0], "*.pyc"))
for filename in pycFiles:
    # check for matching py file
    if not os.path.isfile(filename[:-2]):
        try:
            os.remove(filename)
        except:
            pass  # may not have sufficient privs


def getAllCategories(folderList=()):
    allComps = getAllComponents(folderList)
    # Hardcode some categories to always appear first/last
    firstCats = ['Favorites', 'Stimuli', 'Responses', 'Custom']
    lastCats = ['I/O', 'Other']
    # Start getting categories
    allCats = firstCats
    for name, thisComp in list(allComps.items()):
        for thisCat in thisComp.categories:
            if thisCat not in allCats + lastCats:
                allCats.append(thisCat)
    return allCats + lastCats


def getAllComponents(folderList=(), fetchIcons=True):
    """Get a dictionary of all available components, from the builtins as well
    as all folders in the folderlist.
    User-defined components will override built-ins with the same name.
    """
    if isinstance(folderList, str):
        raise TypeError('folderList should be iterable, not a string')
    components = getComponents(fetchIcons=fetchIcons)  # get the built-ins
    for folder in folderList:
        userComps = getComponents(folder)
        for thisKey in userComps:
            components[thisKey] = userComps[thisKey]

    # add components registered by plugins that have been loaded
    components.update(pluginComponents)

    return components


def getComponents(folder=None, fetchIcons=True):
    """Get a dictionary of available components for the Builder experiments.

    If folder is None then the built-in components will be imported and
    returned, otherwise the components found in the folder provided will be.

    Changed v1.84.00:
    The Builder preference "components folders" should be of the form:
    `/.../.../compts`. This is unchanged from previously, and allows for
    backwards compatibility.

    New as of v1.84: A slightly different directory structure is needed. An
    existing directory will be automatically upgraded if the previous one
    was not empty. However, files starting with '_' will be skipped, as will any
    directories. You will need to manually move those into the new
    directory (from the old .../compts/ into the new .../compts/compts/).

    As of v1.84, a components path needs directory structure `/.../.../compts/compts`.
    That is, the path should end with the name repeated. (It does not need to
    be 'compts' literally.)
    The .py and .png files for a component should all go in this directory.
    (Previously, files were in `/.../.../compts`.) In addition, the directory
    must contain a python init file, ` /.../.../compts/compts/__init__.py`,
    to allow it to be treated as a module in python so that the components
    can be imported. For this reason, the file name for a component
    cannot begin with a number; it must be a legal python name.

    The code within the component.py file itself must use absolute paths for
    importing from psychopy:
       `from psychopy.experiment.components import BaseComponent, Param`
    """
    if folder is None:
        pth = folder = dirname(__file__)
        pkg = 'psychopy.experiment.components'
    else:
        # default shared location is often not actually a folder
        if not os.path.isdir(folder):
            return {}
        pth = folder = folder.rstrip(os.sep)
        pkg = os.path.basename(folder)
        if not folder.endswith(join(pkg, pkg)):
            folder = os.path.join(folder, pkg)

        # update the old style directory (v1.83.03) to the new style
        # try to retain backwards compatibility: copy files, not move them
        # ideally hard link them, but permissions fail on windows
        if not os.path.isdir(folder):
            files = [f for f in glob.glob(join(pth, '*'))
                     if not os.path.isdir(f) and
                     not f[0] in '_0123456789']
            if files:
                os.mkdir(folder)
                with open(join(folder, '__init__.py'), 'a') as fileh:
                    fileh.write('')
                for f in files:
                    if f.startswith('_'):
                        continue
                    shutil.copy(f, folder)
    if not pth in os.sys.path:
        os.sys.path.insert(0, pth)

    components = {}

    # go through components in directory
    cfiles = glob.glob(os.path.join(folder, '*.py'))  # old-style: just comp.py
    # new-style: directories w/ __init__.py
    dfiles = [d for d in os.listdir(folder)
              if os.path.isdir(os.path.join(folder, d))]
    for cmpfile in cfiles + dfiles:
        cmpfile = os.path.split(cmpfile)[1]
        if cmpfile[0] in '_0123456789':  # __init__.py, _base.py, leading digit
            continue
        # can't use imp - breaks py2app:
        # module = imp.load_source(file[:-3], fullPath)
        # v1.83.00 used exec(implicit-relative), no go for python3:
        # exec('import %s as module' % file[:-3])
        # importlib.import_module eases 2.7 -> 3.x migration
        if cmpfile.endswith('.py'):
            explicit_rel_path = pkg + '.' + cmpfile[:-3]
        else:
            explicit_rel_path = pkg + '.' + cmpfile
        try:
            module = import_module(explicit_rel_path, package=pkg)
        except ImportError:
            logging.error(
                'Failed to load component package `{}`. Does it have a '
                '`__init__.py`?'.format(cmpfile))
            continue  # not a valid module (no __init__.py?)
            
        # check for orphaned pyc files (__file__ is not a .py file)
        if hasattr(module, '__file__'):
            if not module.__file__:
                # with Py3, orphans have a __pycharm__ folder but no file
                continue
            elif module.__file__.endswith('.pyc'):
                # with Py2, orphans have a xxxxx.pyc file
                if not os.path.isfile(module.__file__[:-1]):
                    continue  # looks like an orphaned pyc file
        # give a default category
        if not hasattr(module, 'categories'):
            module.categories = ['Custom']
        # check if module contains a component
        for attrib in dir(module):
            name = None
            # fetch the attribs that end with 'Component'
            if (attrib.endswith('omponent') and
                    attrib not in excludeComponents):
                name = attrib
                components[attrib] = getattr(module, attrib)

                # skip if this class was imported, not defined here
                if module.__name__ != components[attrib].__module__:
                    continue  # class was defined in different module

                if hasattr(module, 'tooltip'):
                    tooltips[name] = module.tooltip
                if hasattr(components[attrib], 'iconFile'):
                    iconFiles[name] = components[attrib].iconFile
                # assign the module categories to the Component
                if not hasattr(components[attrib], 'categories'):
                    components[attrib].categories = ['Custom']
    return components



def getInitVals(params, target="PsychoPy"):
    """Works out a suitable initial value for a parameter (e.g. to go into the
    __init__ of a stimulus object, avoiding using a variable name if possible
    """
    inits = copy.deepcopy(params)
    for name in params:
        if target == "PsychoJS":
            # convert (0,0.5) to [0,0.5] but don't convert "rand()" to "rand[]" and don't convert text
            valStr = str(inits[name].val).strip()
            if valStr.startswith("(") and valStr.endswith(")") and name != 'text':
                inits[name].val = py2js.expression2js(inits[name].val)
            # filenames (e.g. for image) need to be loaded from resources
            if name in ["sound"]:
                val = str(inits[name].val)
                if val not in [None, 'None', 'none', '']:
                    inits[name].val = ("psychoJS.resourceManager.getResource({})"
                                       .format(inits[name]))
                    inits[name].valType = 'code'

        if not hasattr(inits[name], 'updates'):  # might be settings parameter instead
            continue

        # value should be None (as code)
        elif inits[name].val in [None, 'None', 'none', '']:
            if name in ['text']:
                inits[name].val = None
                inits[name].valType = 'extendedStr'
            else:
                inits[name].val = 'None'
                inits[name].valType = 'code'

        # is constant so don't touch the parameter value
        elif inits[name].updates in ['constant', None, 'None']:
            continue  # things that are constant don't need handling

        # is changing so work out a reasonable default
        elif name in ['pos', 'fieldPos']:
            inits[name].val = '[0,0]'
            inits[name].valType = 'code'
        elif name in ['color', 'foreColor', 'borderColor', 'lineColor', 'fillColor']:
            inits[name].val = 'white'
            inits[name].valType = 'str'
        elif name in ['ori', 'sf', 'size', 'height', 'letterHeight', 'lineWidth',
                      'phase', 'opacity',
                      'volume',  # sounds
                      'coherence', 'nDots', 'fieldSize', 'dotSize', 'dotLife',
                      'dir', 'speed',
                      'contrast', 'moddepth', 'envori', 'envphase', 'envsf',
                      'noiseClip', 'noiseBWO', 'noiseFilterUpper', 'noiseFilterLower',
                      'noiseBaseSf', 'noiseBW', 'noiseElementSize', 'noiseFilterOrder',
                      'noiseFractalPower']:
            inits[name].val = "1.0"
            inits[name].valType = 'code'
        elif name in ['image', 'mask', 'envelope', 'carrier']:
            inits[name].val = "sin"
            inits[name].valType = 'str'
        elif name == 'texture resolution':
            inits[name].val = "128"
            inits[name].valType = 'code'
        elif name == 'colorSpace':
            inits[name].val = "rgb"
            inits[name].valType = 'str'
        elif name == 'font':
            inits[name].val = "Arial"
            inits[name].valType = 'str'
        elif name == 'units':
            inits[name].val = "norm"
            inits[name].valType = 'str'
        elif name == 'text':
            inits[name].val = ""
            inits[name].valType = 'str'
        elif name == 'flip':
            inits[name].val = ""
            inits[name].valType = 'str'
        elif name == 'sound':
            inits[name].val = "A"
            inits[name].valType = 'str'
        elif name == 'blendmode':
            inits[name].val = "avg"
            inits[name].valType = 'str'
        elif name == 'beat':
            inits[name].val = "False"
            inits[name].valType = 'str'
        elif name == 'noiseImage':
            inits[name].val = "None"
            inits[name].valType = 'str'
        elif name == 'noiseType':
            inits[name].val = 'Binary'
            inits[name].valType = 'str'
        elif name == 'marker_label':
            inits[name].val = 'Label'
            inits[name].valType = 'str'
        elif name == 'marker_value':
            inits[name].val = 'Value'
            inits[name].valType = 'str'
        elif name == 'buttonRequired':
            inits[name].val = "True"
            inits[name].valType = 'code'
        elif name == 'vertices':
            inits[name].val = "[[-0.5,-0.5], [-0.5, 0.5], [0.5, 0.5], [0.5, -0.5]]"
            inits[name].valType = 'code'
        elif name == 'movie':
            inits[name].val = 'None'
            inits[name].valType = 'code'
        else:
            print("I don't know the appropriate default value for a '%s' "
                  "parameter. Please email the mailing list about this error" %
                  name)

    return inits

tooltips = {}
iconFiles = {}



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from ctypes import cdll

from . import MouseDevice
from .. import Keyboard, Computer, Device, xlib
from ...constants import MouseConstants
from ...errors import print2err, printExceptionDetailsToStdErr

currentSec = Computer.getTime


class Mouse(MouseDevice):
    """The Mouse class and related events represent a standard computer mouse
    device and the events a standard mouse can produce.

    Mouse position data is mapped to the coordinate space defined in the
    ioHub configuration file for the Display.

    """

    _xdll = None
    _xfixsdll = None
    _xdisplay = None
    _xscreen_count = None

    __slots__ = ['_cursorVisible']

    def __init__(self, *args, **kwargs):
        MouseDevice.__init__(self, *args, **kwargs['dconfig'])

        self._cursorVisible = True

        if Mouse._xdll is None:
            try:
                Mouse._xdll = cdll.LoadLibrary('libX11.so')
                try:
                    # should use linux cmd:
                    # find /usr/lib -name libXfixes.so\*
                    # to find full path to the lib (if it exists)
                    #
                    Mouse._xfixsdll = cdll.LoadLibrary('libXfixes.so')
                except Exception:
                    try:
                        Mouse._xfixsdll = cdll.LoadLibrary(
                            'libXfixes.so.3.1.0')
                    except Exception:
                        print2err(
                            'ERROR: Mouse._xfixsdll is None. libXfixes.so could not be found')
            except Exception:
                print2err(
                    'ERROR: Mouse._xdll is None. libX11.so could not be found')

        Mouse._xdisplay = xlib.XOpenDisplay(None)
        Mouse._xscreen_count = xlib.XScreenCount(Mouse._xdisplay)

        if Mouse._xfixsdll and self._xdll and self._display_device and self._display_device._xwindow is None:
            self._display_device._xwindow = self._xdll.XRootWindow(
                Mouse._xdisplay, self._display_device.getIndex())


    def _initialMousePos(self):
        """If getPosition is called prior to any mouse events being received,
        this method gets the current system cursor pos.
        TODO: Implement Linux version
        """
        if self._position is None:
            self._position = 0.0, 0.0
            self._lastPosition = 0.0, 0.0

    def _nativeSetMousePos(self, px, py):
        Mouse._xdll.XWarpPointer(
            Mouse._xdisplay,
            None,
            self._display_device._xwindow,
            0,
            0,
            0,
            0,
            int(px),
            int(py))
        Mouse._xdll.XFlush(Mouse._xdisplay)

    def _nativeEventCallback(self, event):
        try:
            if self.isReportingEvents():
                logged_time = currentSec()
                event_array = event[0]

                psychowins = self._iohub_server._psychopy_windows.keys()
                report_all = self.getConfiguration().get('report_system_wide_events', True)
                if report_all is False and psychowins and event_array[-1] not in psychowins:
                    return True

                event_array[3] = Device._getNextEventID()

                display_index = self.getDisplayIndexForMousePosition((event_array[15], event_array[16]))
                if display_index == -1:
                    if self._last_display_index is not None:
                        display_index = self._last_display_index
                    else:
                        # Do not report event to iohub if it does not map to a display
                        # ?? Can this ever actually happen ??
                        return True

                enable_multi_window = self.getConfiguration().get('enable_multi_window', False)
                if enable_multi_window is False:
                    # convert mouse position to psychopy window coord space
                    display_index = self._display_device.getIndex()
                    x, y = self._display_device._pixel2DisplayCoord(event_array[15], event_array[16], display_index)
                    event_array[15] = x
                    event_array[16] = y
                else:
                    wid, wx, wy = self._desktopToWindowPos((event_array[15], event_array[16]))
                    if wid:
                        wx, wy = self._pix2windowUnits(wid, (wx, wy))
                        event_array[15], event_array[16] = x, y = wx, wy
                        event_array[-1] = wid
                    else:
                        event_array[-1] = 0
                        x = event_array[15]
                        y = event_array[16]

                event_array[-2] = Keyboard._modifier_value
                self._lastPosition = self._position
                self._position = x, y

                event_array[11] = display_index
                self._last_display_index = self._display_index
                self._display_index = display_index

                bstate = event_array[-11]
                bnum = event_array[-10]

                if bnum is not MouseConstants.MOUSE_BUTTON_NONE:
                    self.activeButtons[bnum] = int(bstate)

                self._scrollPositionY = event_array[-3]

                self._addNativeEventToBuffer(event_array)

                self._last_callback_time = logged_time
        except Exception:
            printExceptionDetailsToStdErr()
        return True

    def _getIOHubEventObject(self, native_event_data):
        return native_event_data

    def _close(self):
        if Mouse._xdll:
            #if Mouse._xfixsdll and self._nativeGetSystemCursorVisibility() is False:
            #    Mouse._xfixsdll.XFixesShowCursor(
            #        Mouse._xdisplay, self._display_device._xwindow)
            Mouse._xdll.XCloseDisplay(Mouse._xdisplay)
            Mouse._xdll = None
            Mouse._xfixsdll = None
            Mouse._xdisplay = None
            Mouse._xscreen_count = None

            try:
                self._display_device._xwindow = None
            except Exception:
                pass



from copy import deepcopy

from .. import BaseStandaloneRoutine
from psychopy.localization import _translate
from psychopy.experiment import Param
from pathlib import Path
from psychopy.alerts import alert

positions = ['THREE_POINTS', 'FIVE_POINTS', 'NINE_POINTS', "THIRTEEN_POINTS", "SEVENTEEN_POINTS"]

class EyetrackerValidationRoutine(BaseStandaloneRoutine):
    categories = ['Eyetracking']
    targets = ["PsychoPy"]
    iconFile = Path(__file__).parent / "eyetracker_valid.png"
    tooltip = _translate("Validation routine for eyetrackers")
    beta = True

    def __init__(self, exp, name='validation',
                 showCursor=True, cursorFillColor="green",
                 innerFillColor='green', innerBorderColor='black', innerBorderWidth=2, innerRadius=0.0035,
                 fillColor='', borderColor="black", borderWidth=2, outerRadius=0.01,
                 colorSpace="rgb", units='from exp settings', textColor="auto",
                 randomisePos=True, targetLayout="NINE_POINTS", targetPositions="NINE_POINTS",
                 progressMode="time", targetDur=1.5, expandDur=1, expandScale=1.5,
                 movementAnimation=True, movementDur=1.0, targetDelay=1.0,
                 saveAsImg=False, showResults=True,
                 disabled=False
                 ):
        # Initialise base routine
        BaseStandaloneRoutine.__init__(self, exp, name=name, disabled=disabled)

        self.exp.requirePsychopyLibs(['iohub', 'hardware'])

        # Define relationships
        self.depends = [  # allows params to turn each other off/on
            # Only enable positions if targetLayout is custom
            {"dependsOn": "targetLayout",  # must be param name
             "condition": "=='custom...'",  # val to check for
             "param": "positions",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             },
        ]

        # Basic params
        del self.params['stopVal']
        del self.params['stopType']

        self.order += [
            "targetLayout",
            "targetPositions",
            "randomisePos",
            "showCursor",
            "cursorFillColor",
            "textColor"
        ]

        self.params['targetLayout'] = Param(targetLayout,
                                            valType='str', inputType="choice", categ='Basic',
                                            allowedVals=positions + ["CUSTOM..."],
                                            hint=_translate("Pre-defined target layouts"),
                                            label=_translate("Target Layout"))

        self.depends.append(
            {"dependsOn": "targetLayout",  # must be param name
             "condition": "not in {}".format(positions),  # val to check for
             "param": "targetPositions",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['targetPositions'] = Param(targetPositions,
                                               valType='list', inputType="single", categ='Basic',
                                               hint=_translate(
                                                   "List of positions (x, y) at which the target can appear"),
                                               label=_translate("Target Positions"))

        self.params['randomisePos'] = Param(randomisePos,
                                            valType='bool', inputType="bool", categ='Basic',
                                            hint=_translate("Should the order of target positions be randomised?"),
                                            label=_translate("Randomise Target Positions"))

        self.params['cursorFillColor'] = Param(cursorFillColor,
                                               valType="color", inputType="color", categ="Basic",
                                               hint=_translate("Fill color of the gaze cursor"),
                                               label=_translate("Gaze Cursor Color"))

        self.params['textColor'] = Param(textColor,
                                               valType="color", inputType="color", categ="Basic",
                                               hint=_translate("Color of text used in validation procedure."),
                                               label=_translate("Text Color"))

        # Target Params
        self.order += [
            "targetStyle",
            "fillColor",
            "borderColor",
            "innerFillColor",
            "innerBorderColor",
            "colorSpace",
            "borderWidth",
            "innerBorderWidth",
            "outerRadius",
            "innerRadius",
        ]

        self.params['innerFillColor'] = Param(innerFillColor,
                                              valType='color', inputType="color", categ='Target',
                                              hint=_translate("Fill color of the inner part of the target"),
                                              label=_translate("Inner Fill Color"))

        self.params['innerBorderColor'] = Param(innerBorderColor,
                                                valType='color', inputType="color", categ='Target',
                                                hint=_translate("Border color of the inner part of the target"),
                                                label=_translate("Inner Border Color"))

        self.params['fillColor'] = Param(fillColor,
                                         valType='color', inputType="color", categ='Target',
                                         hint=_translate("Fill color of the outer part of the target"),
                                         label=_translate("Outer Fill Color"))

        self.params['borderColor'] = Param(borderColor,
                                           valType='color', inputType="color", categ='Target',
                                           hint=_translate("Border color of the outer part of the target"),
                                           label=_translate("Outer Border Color"))

        self.params['colorSpace'] = Param(colorSpace,
                                          valType='str', inputType="choice", categ='Target',
                                          allowedVals=['rgb', 'dkl', 'lms', 'hsv'],
                                          hint=_translate(
                                              "In what format (color space) have you specified the colors? (rgb, dkl, lms, hsv)"),
                                          label=_translate("Color Space"))

        self.params['borderWidth'] = Param(borderWidth,
                                           valType='num', inputType="single", categ='Target',
                                           hint=_translate("Width of the line around the outer part of the target"),
                                           label=_translate("Outer Border Width"))

        self.params['innerBorderWidth'] = Param(innerBorderWidth,
                                                valType='num', inputType="single", categ='Target',
                                                hint=_translate(
                                                    "Width of the line around the inner part of the target"),
                                                label=_translate("Inner Border Width"))

        self.params['outerRadius'] = Param(outerRadius,
                                           valType='num', inputType="single", categ='Target',
                                           hint=_translate("Size (radius) of the outer part of the target"),
                                           label=_translate("Outer Radius"))

        self.params['innerRadius'] = Param(innerRadius,
                                           valType='num', inputType="single", categ='Target',
                                           hint=_translate("Size (radius) of the inner part of the target"),
                                           label=_translate("Inner Radius"))

        self.params['units'] = Param(units,
                                     valType='str', inputType="choice", categ='Target',
                                     allowedVals=['from exp settings'], direct=False,
                                     hint=_translate("Units of dimensions for this stimulus"),
                                     label=_translate("Spatial Units"))

        # Animation Params
        self.order += [
            "progressMode",
            "targetDur",
            "expandDur",
            "expandScale",
            "movementAnimation",
            "movementDur",
            "targetDelay"
        ]

        self.params['progressMode'] = Param(progressMode,
                                            valType="str", inputType="choice", categ="Animation",
                                            allowedVals=["space key", "time"],
                                            hint=_translate("Should the target move to the next position after a "
                                                            "keypress or after an amount of time?"),
                                            label=_translate("Progress Mode"))

        self.depends.append(
            {"dependsOn": "progressMode",  # must be param name
             "condition": "in ['time', 'either']",  # val to check for
             "param": "targetDur",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['targetDur'] = Param(targetDur,
                                            valType='num', inputType="single", categ='Animation',
                                            hint=_translate(
                                                "Time limit (s) after which progress to next position"),
                                            label=_translate("Target Duration"))

        self.depends.append(
            {"dependsOn": "progressMode",  # must be param name
             "condition": "in ['space key', 'either']",  # val to check for
             "param": "expandDur",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['expandDur'] = Param(expandDur,
                                           valType='num', inputType="single", categ='Animation',
                                           hint=_translate(
                                               "Duration of the target expand/contract animation"),
                                           label=_translate("Expand / Contract Duration"))

        self.params['expandScale'] = Param(expandScale,
                                           valType='num', inputType="single", categ='Animation',
                                           hint=_translate("How many times bigger than its size the target grows"),
                                           label=_translate("Expand Scale"))

        self.params['movementAnimation'] = Param(movementAnimation,
                                           valType='bool', inputType="bool", categ='Animation',
                                           hint=_translate("Enable / disable animations as target stim changes position"),
                                           label=_translate("Animate Position Changes"))

        self.depends.append(
            {"dependsOn": "movementAnimation",  # must be param name
             "condition": "== True",  # val to check for
             "param": "movementDur",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['movementDur'] = Param(movementDur,
                                           valType='num', inputType="single", categ='Animation',
                                           hint=_translate(
                                               "Duration of the animation during position changes."),
                                           label=_translate("Movement Duration"))

        self.depends.append(
            {"dependsOn": "movementAnimation",  # must be param name
             "condition": "== False",  # val to check for
             "param": "targetDelay",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['targetDelay'] = Param(targetDelay,
                                           valType='num', inputType="single", categ='Animation',
                                           hint=_translate(
                                               "Duration of the delay between positions."),
                                           label=_translate("Target Delay"))

        # Data params
        self.order += [
            "saveAsImg",
            "showResults",
        ]

        self.params['saveAsImg'] = Param(saveAsImg,
            valType='bool', inputType="bool", categ='Data',
            hint=_translate(
                "Save results as an image"),
            label=_translate("Save As Image"))

        self.params['showResults'] = Param(showResults,
            valType='bool', inputType="bool", categ='Data',
            hint=_translate(
                "Show a screen with results after completion?"),
            label=_translate("Show Results Screen"))

    def writeMainCode(self, buff):
        # Alert user if eyetracking isn't setup
        if self.exp.eyetracking == "None":
            alert(code=4505)

        # Get inits
        inits = deepcopy(self.params)
        # Code-ify 'from exp settings'
        if inits['units'].val == 'from exp settings':
            inits['units'].val = None
        # Synonymise expand dur and target dur
        if inits['progressMode'].val == 'time':
            inits['expandDur'] = inits['targetDur']
        if inits['progressMode'].val == 'space key':
            inits['targetDur'] = inits['expandDur']
        # Synonymise movement dur and target delay
        if inits['movementAnimation'].val:
            inits['targetDelay'] = inits['movementDur']
        else:
            inits['movementDur'] = inits['targetDelay']
        # Convert progress mode to ioHub format
        if inits['progressMode'].val == 'space key':
            inits['progressKey'] = "'space'"
        else:
            inits['progressKey'] = "None"
        # If positions are preset, override param value
        if inits['targetLayout'].val in positions:
            inits['targetPositions'].val = inits['targetLayout'].val
            inits['targetPositions'].valType = 'str'

        BaseStandaloneRoutine.writeMainCode(self, buff)

        # Make target
        code = (
            "# define target for %(name)s\n"
            "%(name)sTarget = visual.TargetStim(win, \n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                "name='%(name)sTarget',\n"
                "radius=%(outerRadius)s, fillColor=%(fillColor)s, borderColor=%(borderColor)s, lineWidth=%(borderWidth)s,\n"
                "innerRadius=%(innerRadius)s, innerFillColor=%(innerFillColor)s, innerBorderColor=%(innerBorderColor)s, innerLineWidth=%(innerBorderWidth)s,\n"
                "colorSpace=%(colorSpace)s, units=%(units)s\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            ")"
        )
        buff.writeIndentedLines(code % inits)

        # Make validation object
        code = (
            "# define parameters for %(name)s\n"
            "%(name)s = iohub.ValidationProcedure(win,\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)

        code = (
                "target=%(name)sTarget,\n"
                "gaze_cursor=%(cursorFillColor)s, \n"
                "positions=%(targetPositions)s, randomize_positions=%(randomisePos)s,\n"
                "expand_scale=%(expandScale)s, target_duration=%(targetDur)s,\n"
                "enable_position_animation=%(movementAnimation)s, target_delay=%(targetDelay)s,\n"
                "progress_on_key=%(progressKey)s, text_color=%(textColor)s,\n"
                "show_results_screen=%(showResults)s, save_results_screen=%(saveAsImg)s,\n"
                "color_space=%(colorSpace)s, unit_type=%(units)s\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            ")\n"
        )
        buff.writeIndentedLines(code % inits)
        # Run
        code = (
            "# run %(name)s\n"
            "%(name)s.run()\n"
            "# clear any keypresses from during %(name)s so they don't interfere with the experiment\n"
            "defaultKeyboard.clearEvents()\n"
        )
        buff.writeIndentedLines(code % inits)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A stimulus class for playing movies (mpeg, avi, etc...) in PsychoPy using a
local installation of VLC media player (https://www.videolan.org/).
"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
#
# VlcMovieStim originally contributed by Dan Fitch, April 2019. The `MovieStim2`
# class was taken and rewritten to use only VLC.
#

import os
import sys
import threading
import ctypes
import weakref

from psychopy import core, logging
from psychopy.tools.attributetools import logAttrib, setAttribute
from psychopy.tools.filetools import pathToString
from psychopy.visual.basevisual import BaseVisualStim, ContainerMixin
from psychopy.constants import FINISHED, NOT_STARTED, PAUSED, PLAYING, STOPPED

import numpy
import pyglet
pyglet.options['debug_gl'] = False
GL = pyglet.gl

try:
    # check if the lib can be loaded
    import vlc
    haveVLC = True
except Exception as err:
    haveVLC = False
    # store the error but only raise it if the
    if "wrong architecture" in str(err):
        msg = ("Failed to import `vlc` module required by `vlcmoviestim`.\n"
               "You're using %i-bit python. Is your VLC install the same?"
               % 64 if sys.maxsize == 2 ** 64 else 32)
        _vlcImportErr = OSError(msg)
    else:
        _vlcImportErr = err

# flip time, and time since last movie frame flip will be printed
reportNDroppedFrames = 10


class VlcMovieStim(BaseVisualStim, ContainerMixin):
    """A stimulus class for playing movies in various formats (mpeg, avi,
    etc...) in PsychoPy using the VLC media player as a decoder.

    This movie class is very efficient and better suited for playing
    high-resolution videos (720p+) than the other movie classes. However, audio
    is only played using the default output device. This may be adequate for
    most applications where the user is not concerned about precision audio
    onset times.

    The VLC media player (https://www.videolan.org/) must be installed on the
    machine running PsychoPy to use this class. Make certain that the version
    of VLC installed matches the architecture of the Python interpreter hosting
    PsychoPy.

    Parameters
    ----------
    win : :class:`~psychopy.visual.Window`
        Window the video is being drawn to.
    filename : str
        Name of the file or stream URL to play. If an empty string, no file will
        be loaded on initialization but can be set later.
    units : str
        Units to use when sizing the video frame on the window, affects how
        `size` is interpreted.
    size : ArrayLike or None
        Size of the video frame on the window in `units`. If `None`, the native
        size of the video will be used.
    flipVert : bool
        If `True` then the movie will be top-bottom flipped.
    flipHoriz : bool
        If `True` then the movie will be right-left flipped.
    volume : int or float
        If specifying an `int` the nominal level is 100, and 0 is silence. If a
        `float`, values between 0 and 1 may be used.
    loop : bool
        Whether to start the movie over from the beginning if draw is called and
        the movie is done. Default is `False.
    autoStart : bool
        Automatically begin playback of the video when `flip()` is called.

    Notes
    -----
    * You may see error messages in your log output from VLC (e.g.,
      `get_buffer() failed`, `no frame!`, etc.) after shutting down. These
      errors originate from the decoder and can be safely ignored.

    """
    def __init__(self, win,
                 filename="",
                 units='pix',
                 size=None,
                 pos=(0.0, 0.0),
                 ori=0.0,
                 flipVert=False,
                 flipHoriz=False,
                 color=(1.0, 1.0, 1.0),  # remove?
                 colorSpace='rgb',
                 opacity=1.0,
                 volume=1.0,
                 name='',
                 loop=False,
                 autoLog=True,
                 depth=0.0,
                 noAudio=False,
                 interpolate=True,
                 autoStart=True):
        # check if we have the VLC lib
        if not haveVLC:
            raise _vlcImportErr
        # what local vars are defined (these are the init params) for use
        # by __repr__
        self._initParams = dir()
        self._initParams.remove('self')
        super(VlcMovieStim, self).__init__(win, units=units, name=name,
                                         autoLog=False)
        # check for pyglet
        if win.winType != 'pyglet':
            logging.error('Movie stimuli can only be used with a pyglet window')
            core.quit()

        # drawing stuff
        self.flipVert = flipVert
        self.flipHoriz = flipHoriz
        self.pos = numpy.asarray(pos, float)

        # original size to keep BaseVisualStim happy
        self._origSize = numpy.asarray((128, 128,), float)

        # Defer setting size until after the video is loaded to use it's native
        # size instead of the one set by the user.
        self._useFrameSizeFromVideo = size is None
        if not self._useFrameSizeFromVideo:
            self.size = numpy.asarray(size, float)

        self.depth = depth
        self.opacity = float(opacity)

        # playback stuff
        self._filename = pathToString(filename)
        self._volume = volume
        self._noAudio = noAudio  # cannot be changed
        self._currentFrame = -1
        self._loopCount = 0
        self.loop = loop

        # video pixel and texture buffer variables, setup later
        self.interpolate = interpolate  # use setter
        self._textureId = GL.GLuint()
        self._pixbuffId = GL.GLuint()

        # VLC related attributes
        self._instance = None
        self._player = None
        self._manager = None
        self._stream = None
        self._videoWidth = 0
        self._videoHeight = 0
        self._frameRate = 0.0
        self._vlcInitialized = False
        self._pixelLock = threading.Lock()  # semaphore for VLC pixel transfer
        self._framePixelBuffer = None  # buffer pointer to draw to
        self._videoFrameBufferSize = None
        self._streamEnded = False

        # spawn a VLC instance for this class instance
        # self._createVLCInstance()

        # load a movie if provided
        if self._filename:
            self.loadMovie(self._filename)

        self.setVolume(volume)
        self.nDroppedFrames = 0
        self._autoStart = autoStart

        self.ori = ori
        # set autoLog (now that params have been initialised)
        self.autoLog = autoLog
        if autoLog:
            logging.exp("Created {} = {}".format(self.name, self))

    @property
    def filename(self):
        """File name for the loaded video (`str`)."""
        return self._filename

    @filename.setter
    def filename(self, value):
        self.loadMovie(value)

    @property
    def autoStart(self):
        """Start playback when `.draw()` is called (`bool`)."""
        return self._autoStart

    @autoStart.setter
    def autoStart(self, value):
        self._autoStart = bool(value)

    def setMovie(self, filename, log=True):
        """See `~MovieStim.loadMovie` (the functions are identical).

        This form is provided for syntactic consistency with other visual
        stimuli.
        """
        self.loadMovie(filename, log=log)

    def loadMovie(self, filename, log=True):
        """Load a movie from file

        Parameters
        ----------
        filename : str
            The name of the file or URL, including path if necessary.
        log : bool
            Log this event.

        Notes
        -----
        * Due to VLC oddness, `.duration` is not correct until the movie starts
          playing.

        """
        self._filename = pathToString(filename)

        # open the media using a new player
        self._openMedia()

        self.status = NOT_STARTED
        logAttrib(self, log, 'movie', filename)

    def _createVLCInstance(self):
        """Internal method to create a new VLC instance.

        Raises an error if an instance is already spawned and hasn't been
        released.
        """
        logging.debug("Spawning new VLC instance ...")

        # Creating VLC instances is slow, so we want to create once per class
        # instantiation and reuse it as much as possible. Stopping and starting
        # instances too frequently can result in an errors and affect the
        # stability of the system.

        if self._instance is not None:
            self._releaseVLCInstance()
            # errmsg = ("Attempted to create another VLC instance without "
            #           "releasing the previous one first!")
            # logging.fatal(errmsg, obj=self)
            # logging.flush()
            # raise RuntimeError(errmsg)

        # Using "--quiet" here is just sweeping anything VLC pukes out under the
        # rug. Most of the time the errors only look scary but can be ignored.
        params = " ".join(
            ["--no-audio" if self._noAudio else "", "--sout-keep", "--quiet"])
        self._instance = vlc.Instance(params)

        # used to capture log messages from VLC
        self._instance.log_set(vlcLogCallback, None)

        # create a new player object, reusable by by just changing the stream
        self._player = self._instance.media_player_new()

        # setup the event manager
        self._manager = self._player.event_manager()
        # self._manager.event_attach(
        #     vlc.EventType.MediaPlayerTimeChanged, vlcMediaEventCallback,
        #     weakref.ref(self), self._player)
        self._manager.event_attach(
            vlc.EventType.MediaPlayerEndReached, vlcMediaEventCallback,
            weakref.ref(self), self._player)

        self._vlcInitialized = True

        logging.debug("VLC instance created.")

    def _releaseVLCInstance(self):
        """Internal method to release a VLC instance. Calling this implicitly
        stops and releases any stream presently loaded and playing.
        """
        self._vlcInitialized = False

        if self._player is not None:
            self._player.stop()

            # shutdown the manager
            self._manager.event_detach(vlc.EventType.MediaPlayerEndReached)
            self._manager = None

            # Doing this here since I figured Python wasn't shutting down due to
            # callbacks remaining bound. Seems to actually fix the problem.
            self._player.video_set_callbacks(None, None, None, None)
            self._player.set_media(None)
            self._player.release()
            self._player = None

            # reset video information
            self._filename = None
            self._videoWidth = self._videoHeight = 0
            self._frameCounter = self._loopCount = 0
            self._frameRate = 0.0
            self._framePixelBuffer = None

        if self._stream is not None:
            self._stream.release()
            self._streamEnded = False
            self._stream = None

        if self._instance is not None:
            self._instance.release()
            self._instance = None

        self.status = STOPPED

    def _openMedia(self, uri=None):
        """Internal method that opens a new stream using `filename`. This will
        close the previous stream. Raises an error if a VLC instance is not
        available.
        """
        # if None, use `filename`
        uri = self.filename if uri is None else uri

        # create a fresh VLC instance
        self._createVLCInstance()

        # raise error if there is no VLC instance
        if self._instance is None:
            errmsg = "Cannot open a stream without a VLC instance started."
            logging.fatal(errmsg, obj=self)
            logging.flush()

        # check if we have a player, stop it if so
        if self._player is None:
            errmsg = "Cannot open a stream without a VLC media player."
            logging.fatal(errmsg, obj=self)
            logging.flush()
        else:
            self._player.stop()  # stop any playback

        # check if the file is valid and readable
        if not os.access(uri, os.R_OK):
            raise RuntimeError('Error: %s file not readable' % uri)

        # close the previous VLC resources if needed
        if self._stream is not None:
            self._stream.release()

        # open the file and create a new stream
        try:
            self._stream = self._instance.media_new(uri)
        except NameError:
            raise ImportError('NameError: %s vs LibVLC %s' % (
                vlc.__version__, vlc.libvlc_get_version()))

        # player object
        self._player.set_media(self._stream)

        # set media related attributes
        self._stream.parse()
        videoWidth, videoHeight = self._player.video_get_size()
        self._videoWidth = videoWidth
        self._videoHeight = videoHeight

        # set the video size
        if self._useFrameSizeFromVideo:
            self.size = self._origSize = numpy.array(
                (self._videoWidth, self._videoHeight), float)

        self._frameRate = self._player.get_fps()
        self._frameCounter = self._loopCount = 0
        self._streamEnded = False

        # uncomment if not using direct GPU write, might be more thread-safe
        self._framePixelBuffer = (
                ctypes.c_ubyte * self._videoWidth * self._videoHeight * 4)()

        # duration unavailable until started
        duration = self._player.get_length()
        logging.info("Video is %ix%i, duration %s, fps %s" % (
            self._videoWidth, self._videoHeight, duration, self._frameRate))
        logging.flush()

        # We assume we can use the RGBA format here
        self._player.video_set_format(
            "RGBA", self._videoWidth, self._videoHeight, self._videoWidth << 2)

        # setup texture buffers before binding the callbacks
        self._setupTextureBuffers()

        # Set callbacks since we have the resources to write to.
        thisInstance = ctypes.cast(
            ctypes.pointer(ctypes.py_object(self)), ctypes.c_void_p)
        self._player.video_set_callbacks(
            vlcLockCallback, vlcUnlockCallback, vlcDisplayCallback,
            thisInstance)

    def _closeMedia(self):
        """Internal method to release the presently loaded stream (if any).
        """
        self._vlcInitialized = False

        if self._player is not None:  # stop any playback
            self._player.stop()
            self._player.set_media(None)

        if self._stream is None:
            return  # nop

        self._stream.release()
        self._stream = self._player = None
        self._useFrameSizeFromVideo = True

        self._releaseVLCInstance()

        # unregister callbacks before freeing buffers
        self._freeBuffers()
        GL.glFlush()

    def _onEos(self):
        """Internal method called when the decoder encounters the end of the
        stream.
        """
        # Do not call the libvlc API in this method, causes a deadlock which
        # freezes PsychoPy. Just set the flag below and the stream will restart
        # on the next `.draw()` call.
        self._streamEnded = True

        if not self.loop:
            self.status = FINISHED

        if self.autoLog:
            self.win.logOnFlip(
                "Set %s finished" % self.name, level=logging.EXP, obj=self)

    def _freeBuffers(self):
        """Free texture and pixel buffers. Call this when tearing down this
        class or if a movie is stopped.
        """
        try:
            # delete buffers and textures if previously created
            if self._pixbuffId.value > 0:
                GL.glDeleteBuffers(1, self._pixbuffId)
                self._pixbuffId = GL.GLuint()

            # delete the old texture if present
            if self._textureId.value > 0:
                GL.glDeleteTextures(1, self._textureId)
                self._textureId = GL.GLuint()

        except TypeError:  # can happen when unloading or shutting down
            pass

    def _setupTextureBuffers(self):
        """Setup texture buffers which hold frame data. This creates a 2D
        RGB texture and pixel buffer. The pixel buffer serves as the store for
        texture color data. Each frame, the pixel buffer memory is mapped and
        frame data is copied over to the GPU from the decoder.

        This is called every time a video file is loaded. The `_freeBuffers`
        method is called in this routine prior to creating new buffers, so it's
        safe to call this right after loading a new movie without having to
        `_freeBuffers` first.

        """
        self._freeBuffers()  # clean up any buffers previously allocated

        # Calculate the total size of the pixel store in bytes needed to hold a
        # single video frame. This value is reused during the pixel upload
        # process. Assumes RGBA color format.
        self._videoFrameBufferSize = \
            self._videoWidth * self._videoHeight * 4 * ctypes.sizeof(GL.GLubyte)

        # Create the pixel buffer object which will serve as the texture memory
        # store. Pixel data will be copied to this buffer each frame.
        GL.glGenBuffers(1, ctypes.byref(self._pixbuffId))
        GL.glBindBuffer(GL.GL_PIXEL_UNPACK_BUFFER, self._pixbuffId)
        GL.glBufferData(
            GL.GL_PIXEL_UNPACK_BUFFER,
            self._videoFrameBufferSize,
            None,
            GL.GL_STREAM_DRAW)  # one-way app -> GL
        GL.glBindBuffer(GL.GL_PIXEL_UNPACK_BUFFER, 0)

        # Create a texture which will hold the data streamed to the pixel
        # buffer. Only one texture needs to be allocated.
        GL.glGenTextures(1, ctypes.byref(self._textureId))
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._textureId)
        GL.glTexImage2D(
            GL.GL_TEXTURE_2D,
            0,
            GL.GL_RGBA8,
            self._videoWidth, self._videoHeight,  # frame dims in pixels
            0,
            GL.GL_RGBA,
            GL.GL_UNSIGNED_BYTE,
            None)
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)  # needs to be 1

        # setup texture filtering
        if self.interpolate:
            texFilter = GL.GL_LINEAR
        else:
            texFilter = GL.GL_NEAREST

        GL.glTexParameteri(
            GL.GL_TEXTURE_2D,
            GL.GL_TEXTURE_MAG_FILTER,
            texFilter)
        GL.glTexParameteri(
            GL.GL_TEXTURE_2D,
            GL.GL_TEXTURE_MIN_FILTER,
            texFilter)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

        GL.glFlush()  # make sure all buffers are ready

    @property
    def frameTexture(self):
        """Texture ID for the current video frame (`GLuint`). You can use this
        as a video texture. However, you must periodically call `updateTexture`
        to keep this up to date.

        """
        return self._textureId

    def _pixelTransfer(self):
        """Internal method which maps the pixel buffer for the video texture
        to client memory, allowing for VLC to directly draw a video frame to it.

        This method is not thread-safe and should never be called without the
        pixel lock semaphore being first set by VLC.

        """
        # bind pixel unpack buffer
        GL.glBindBuffer(GL.GL_PIXEL_UNPACK_BUFFER, self._pixbuffId)

        # Free last storage buffer before mapping and writing new frame data.
        # This allows the GPU to process the extant buffer in VRAM uploaded last
        # cycle without being stalled by the CPU accessing it. Also allows VLC
        # to access the buffer while the previous one is still in use.
        GL.glBufferData(
            GL.GL_PIXEL_UNPACK_BUFFER,
            self._videoFrameBufferSize,
            None,
            GL.GL_STREAM_DRAW)

        # Map the buffer to client memory, `GL_WRITE_ONLY` to tell the driver to
        # optimize for a one-way write operation if it can.
        bufferPtr = GL.glMapBuffer(
            GL.GL_PIXEL_UNPACK_BUFFER,
            GL.GL_WRITE_ONLY)

        # comment to disable direct VLC -> GPU frame write
        # self._framePixelBuffer = bufferPtr

        # uncomment if not using direct GPU write, might provide better thread
        # safety ...
        ctypes.memmove(
            bufferPtr,
            ctypes.byref(self._framePixelBuffer),
            ctypes.sizeof(self._framePixelBuffer))

        # Very important that we unmap the buffer data after copying, but
        # keep the buffer bound for setting the texture.
        GL.glUnmapBuffer(GL.GL_PIXEL_UNPACK_BUFFER)

        # bind the texture in OpenGL
        GL.glEnable(GL.GL_TEXTURE_2D)

        # copy the PBO to the texture
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._textureId)
        GL.glTexSubImage2D(
            GL.GL_TEXTURE_2D, 0, 0, 0,
            self._videoWidth,
            self._videoHeight,
            GL.GL_RGBA,
            GL.GL_UNSIGNED_BYTE,
            0)  # point to the presently bound buffer

        # update texture filtering only if needed
        if self._texFilterNeedsUpdate:
            if self.interpolate:
                texFilter = GL.GL_LINEAR
            else:
                texFilter = GL.GL_NEAREST

            GL.glTexParameteri(
                GL.GL_TEXTURE_2D,
                GL.GL_TEXTURE_MAG_FILTER,
                texFilter)
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D,
                GL.GL_TEXTURE_MIN_FILTER,
                texFilter)

            self._texFilterNeedsUpdate = False

        # important to unbind the PBO
        GL.glBindBuffer(GL.GL_PIXEL_UNPACK_BUFFER, 0)
        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glDisable(GL.GL_TEXTURE_2D)

    def updateTexture(self):
        """Update the video texture buffer to the most recent video frame.
        """
        with self._pixelLock:
            self._pixelTransfer()

    # --------------------------------------------------------------------------
    # Video playback controls and status
    #
    @property
    def isPlaying(self):
        """`True` if the video is presently playing (`bool`)."""
        # Status flags as properties are pretty useful for users since they are
        # self documenting and prevent the user from touching the status flag
        # attribute directly.
        #
        return self.status == PLAYING

    @property
    def isNotStarted(self):
        """`True` if the video has not be started yet (`bool`). This status is
        given after a video is loaded and play has yet to be called."""
        return self.status == NOT_STARTED

    @property
    def isStopped(self):
        """`True` if the video is stopped (`bool`)."""
        return self.status == STOPPED

    @property
    def isPaused(self):
        """`True` if the video is presently paused (`bool`)."""
        return self.status == PAUSED

    @property
    def isFinished(self):
        """`True` if the video is finished (`bool`)."""
        # why is this the same as STOPPED?
        return self.status == FINISHED

    def play(self, log=True):
        """Start or continue a paused movie from current position.

        Parameters
        ----------
        log : bool
            Log the play event.

        Returns
        -------
        int or None
            Frame index playback started at. Should always be `0` if starting at
            the beginning of the video. Returns `None` if the player has not
            been initialized.

        """
        if self.isNotStarted:  # video has not been played yet
            if log and self.autoLog:
                self.win.logOnFlip(
                    "Set %s playing" % self.name,
                    level=logging.EXP,
                    obj=self)

            self._player.play()

        elif self.isPaused:
            self.win.logOnFlip(
                "Resuming playback at position {:.4f}".format(
                    self._player.get_time() / 1000.0),
                level=logging.EXP,
                obj=self)

            if self._player.will_play():
                self._player.play()

        self.status = PLAYING

        return self._currentFrame

    def pause(self, log=True):
        """Pause the current point in the movie.

        Parameters
        ----------
        log : bool
            Log the pause event.

        """
        if self.isPlaying:
            self.status = PAUSED
            player = self._player
            if player and player.can_pause():
                player.pause()
            if log and self.autoLog:
                self.win.logOnFlip("Set %s paused" % self.name,
                                   level=logging.EXP, obj=self)
            return True
        if log and self.autoLog:
            self.win.logOnFlip("Failed Set %s paused" % self.name,
                               level=logging.EXP, obj=self)
        return False

    def stop(self, log=True):
        """Stop the current point in the movie (sound will stop, current frame
        will not advance). Once stopped the movie cannot be restarted - it must
        be loaded again.

        Use `pause()` instead if you may need to restart the movie.

        Parameters
        ----------
        log : bool
            Log the stop event.

        """

        if self._player is None:
            return

        self.status = STOPPED

        if log and self.autoLog:
            self.win.logOnFlip(
                "Set %s stopped" % self.name, level=logging.EXP, obj=self)

        self._releaseVLCInstance()

    def seek(self, timestamp, log=True):
        """Seek to a particular timestamp in the movie.

        Parameters
        ----------
        timestamp : float
            Time in seconds.
        log : bool
            Log the seek event.

        """
        if self.isPlaying or self.isPaused:
            player = self._player
            if player and player.is_seekable():
                # pause while seeking
                player.set_time(int(timestamp * 1000.0))

            if log:
                logAttrib(self, log, 'seek', timestamp)

    def rewind(self, seconds=5):
        """Rewind the video.

        Parameters
        ----------
        seconds : float
            Time in seconds to rewind from the current position. Default is 5
            seconds.

        Returns
        -------
        float
            Timestamp after rewinding the video.

        """
        self.seek(max(self.getCurrentFrameTime() - float(seconds), 0.0))

        # after seeking
        return self.getCurrentFrameTime()

    def fastForward(self, seconds=5):
        """Fast-forward the video.

        Parameters
        ----------
        seconds : float
            Time in seconds to fast forward from the current position. Default
            is 5 seconds.

        Returns
        -------
        float
            Timestamp at new position after fast forwarding the video.

        """
        self.seek(
            min(self.getCurrentFrameTime() + float(seconds), self.duration))

        return self.getCurrentFrameTime()

    def replay(self, autoPlay=True):
        """Replay the movie from the beginning.

        Parameters
        ----------
        autoPlay : bool
            Start playback immediately. If `False`, you must call `play()`
            afterwards to initiate playback.

        Notes
        -----
        * This tears down the current VLC instance and creates a new one.
          Similar to calling `stop()` and `loadMovie()`. Use `seek(0.0)` if you
          would like to restart the movie without reloading.

        """
        lastMovieFile = self._filename
        self.stop()
        self.loadMovie(lastMovieFile)

        if autoPlay:
            self.play()

    # --------------------------------------------------------------------------
    # Volume controls
    #
    @property
    def volume(self):
        """Audio track volume (`int` or `float`). See `setVolume` for more
        information about valid values.

        """
        return self.getVolume()

    @volume.setter
    def volume(self, value):
        self.setVolume(value)

    def setVolume(self, volume):
        """Set the audio track volume.

        Parameters
        ----------
        volume : int or float
            Volume level to set. 0 = mute, 100 = 0 dB. float values between 0.0
            and 1.0 are also accepted, and scaled to an int between 0 and 100.

        """
        if self._player:
            if 0.0 <= volume <= 1.0 and isinstance(volume, float):
                v = int(volume * 100)
            else:
                v = int(volume)
            self._volume = v

            if self._player:
                self._player.audio_set_volume(v)

    def getVolume(self):
        """Returns the current movie audio volume.

        Returns
        -------
        int
            Volume level, 0 is no audio, 100 is max audio volume.

        """
        if self._player:
            self._volume = self._player.audio_get_volume()

        return self._volume

    def increaseVolume(self, amount=10):
        """Increase the volume.

        Parameters
        ----------
        amount : int
            Increase the volume by this amount (percent). This gets added to the
            present volume level. If the value of `amount` and the current
            volume is outside the valid range of 0 to 100, the value will be
            clipped. The default value is 10 (or 10% increase).

        Returns
        -------
        int
            Volume after changed.

        See also
        --------
        getVolume
        setVolume
        decreaseVolume

        Examples
        --------
        Adjust the volume of the current video using key presses::

            # assume `mov` is an instance of this class defined previously
            for key in event.getKeys():
                if key == 'minus':
                    mov.decreaseVolume()
                elif key == 'equals':
                    mov.increaseVolume()

        """
        if not self._player:
            return 0

        self.setVolume(min(max(self.getVolume() + int(amount), 0), 100))

        return self._volume

    def decreaseVolume(self, amount=10):
        """Decrease the volume.

        Parameters
        ----------
        amount : int
            Decrease the volume by this amount (percent). This gets subtracted
            from the present volume level. If the value of `amount` and the
            current volume is outside the valid range of 0 to 100, the value
            will be clipped. The default value is 10 (or 10% decrease).

        Returns
        -------
        int
            Volume after changed.

        See also
        --------
        getVolume
        setVolume
        increaseVolume

        Examples
        --------
        Adjust the volume of the current video using key presses::

            # assume `mov` is an instance of this class defined previously
            for key in event.getKeys():
                if key == 'minus':
                    mov.decreaseVolume()
                elif key == 'equals':
                    mov.increaseVolume()

        """
        if not self._player:
            return 0

        self.setVolume(min(max(self.getVolume() - int(amount), 0), 100))

        return self._volume

    # --------------------------------------------------------------------------
    # Video and playback information
    #
    @property
    def frameIndex(self):
        """Current frame index being displayed (`int`)."""
        return self._currentFrame

    def getCurrentFrameNumber(self):
        """Get the current movie frame number (`int`), same as `frameIndex`.
        """
        return self._frameCounter

    @property
    def percentageComplete(self):
        """Percentage of the video completed (`float`)."""
        return self.getPercentageComplete()

    @property
    def duration(self):
        """Duration of the loaded video in seconds (`float`). Not valid unless
        the video has been started.
        """
        if not self.isNotStarted:
            return self._player.get_length()

        return 0.0

    @property
    def loopCount(self):
        """Number of loops completed since playback started (`int`). This value
        is reset when either `stop` or `loadMovie` is called.
        """
        return self._loopCount

    @property
    def fps(self):
        """Movie frames per second (`float`)."""
        return self.getFPS()

    def getFPS(self):
        """Movie frames per second.

        Returns
        -------
        float
            Nominal number of frames to be displayed per second.

        """
        return self._frameRate

    @property
    def frameTime(self):
        """Current frame time in seconds (`float`)."""
        return self.getCurrentFrameTime()

    def getCurrentFrameTime(self):
        """Get the time that the movie file specified the current video frame as
        having.

        Returns
        -------
        float
            Current video time in seconds.

        """
        if not self._player:
            return 0.0

        return self._player.get_time() / 1000.0

    def getPercentageComplete(self):
        """Provides a value between 0.0 and 100.0, indicating the amount of the
        movie that has been already played.
        """
        return self._player.get_position() * 100.0

    @property
    def videoSize(self):
        """Size of the video `(w, h)` in pixels (`tuple`). Returns `(0, 0)` if
        no video is loaded."""
        if self._stream is not None:
            return self._videoWidth, self._videoHeight

        return 0, 0

    # --------------------------------------------------------------------------
    # Drawing methods and properties
    #
    @property
    def interpolate(self):
        """Enable linear interpolation (`bool').

        If `True` linear filtering will be applied to the video making the image
        less pixelated if scaled. You may leave this off if the native size of
        the video is used.

        """
        return self._interpolate

    @interpolate.setter
    def interpolate(self, value):
        self._interpolate = value
        self._texFilterNeedsUpdate = True

    def setFlipHoriz(self, newVal=True, log=True):
        """If set to True then the movie will be flipped horizontally
        (left-to-right). Note that this is relative to the original, not
        relative to the current state.
        """
        self.flipHoriz = newVal
        logAttrib(self, log, 'flipHoriz')

    def setFlipVert(self, newVal=True, log=True):
        """If set to True then the movie will be flipped vertically
        (top-to-bottom). Note that this is relative to the original, not
        relative to the current state.
        """
        self.flipVert = not newVal
        logAttrib(self, log, 'flipVert')

    def _drawRectangle(self):
        """Draw the frame to the window. This is called by the `draw()` method.
        """
        # make sure that textures are on and GL_TEXTURE0 is activ
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glActiveTexture(GL.GL_TEXTURE0)

        # sets opacity (1, 1, 1 = RGB placeholder)
        GL.glColor4f(1, 1, 1, self.opacity)
        GL.glPushMatrix()
        self.win.setScale('pix')

        # move to centre of stimulus and rotate
        vertsPix = self.verticesPix

        array = (GL.GLfloat * 32)(
            1, 1,  # texture coords
            vertsPix[0, 0], vertsPix[0, 1], 0.,  # vertex
            0, 1,
            vertsPix[1, 0], vertsPix[1, 1], 0.,
            0, 0,
            vertsPix[2, 0], vertsPix[2, 1], 0.,
            1, 0,
            vertsPix[3, 0], vertsPix[3, 1], 0.,
        )
        GL.glPushAttrib(GL.GL_ENABLE_BIT)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self._textureId)
        GL.glPushClientAttrib(GL.GL_CLIENT_VERTEX_ARRAY_BIT)

        # 2D texture array, 3D vertex array
        GL.glInterleavedArrays(GL.GL_T2F_V3F, 0, array)
        GL.glDrawArrays(GL.GL_QUADS, 0, 4)
        GL.glPopClientAttrib()
        GL.glPopAttrib()
        GL.glPopMatrix()

        GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
        GL.glEnable(GL.GL_TEXTURE_2D)

    def draw(self, win=None):
        """Draw the current frame to a particular
        :class:`~psychopy.visual.Window` (or to the default win for this object
        if not specified).

        The current position in the movie will be determined automatically. This
        method should be called on every frame that the movie is meant to
        appear.

        Parameters
        ----------
        win : :class:`~psychopy.visual.Window` or `None`
            Window the video is being drawn to. If `None`, the window specified
            at initialization will be used instead.

        Returns
        -------
        bool
            `True` if the frame was updated this draw call.

        """
        if self.isNotStarted and self._autoStart:
            self.play()
        elif self._streamEnded and self.loop:
            self._loopCount += 1
            self._streamEnded = False
            self.replay()
        elif self.isFinished:
            self.stop()
            return False

        self._selectWindow(self.win if win is None else win)

        # check if we need to pull a new frame this round
        if self._currentFrame == self._frameCounter:

            # update the texture, getting the most recent frame
            self.updateTexture()

            # draw the frame to the window
            self._drawRectangle()

            return False  # frame does not need an update this round

        # Below not called if the frame hasn't advanced yet ------

        # update the current frame
        self._currentFrame = self._frameCounter

        # update the texture, getting the most recent frame
        self.updateTexture()

        # draw the frame to the window
        self._drawRectangle()

        # token gesture for existing code, we handle this logic internally now
        return True

    def setAutoDraw(self, val, log=None):
        """Add or remove a stimulus from the list of stimuli that will be
        automatically drawn on each flip

        :parameters:
            - val: True/False
                True to add the stimulus to the draw list, False to remove it
        """
        if val:
            self.play(log=False)  # set to play in case stopped
        else:
            self.pause(log=False)
        # add to drawing list and update status
        setAttribute(self, 'autoDraw', val, log)

    def __del__(self):
        try:
            if hasattr(self, '_player'):
                # false if crashed before creating instance
                self._releaseVLCInstance()
        except (ImportError, ModuleNotFoundError, TypeError):
            pass  # has probably been garbage-collected already


# ------------------------------------------------------------------------------
# Callback functions for `libvlc`
#
# WARNING: Due to a limitation of libvlc, you cannot call the API from within
# the callbacks below. Doing so will result in a deadlock that stalls the
# application. This applies to any method in the `VloMovieStim` class being
# called within a callback too. They cannot have any libvlc calls anywhere in
# the call stack.
#

@vlc.CallbackDecorators.VideoLockCb
def vlcLockCallback(user_data, planes):
    """Callback invoked when VLC has new texture data.
    """
    # Need to catch errors caused if a NULL object is passed. Not sure why this
    # happens but it might be due to the callback being invoked before the
    # movie stim class is fully realized. This may happen since the VLC library
    # operates in its own thread and may be processing frames before we are
    # ready to use them. This `try-except` structure is present in all callback
    # functions here that pass `user_data` which is a C pointer to the stim
    # object. Right now, these functions just return when we get a NULL object.
    #
    try:
        cls = ctypes.cast(
            user_data, ctypes.POINTER(ctypes.py_object)).contents.value
    except (ValueError, TypeError):
        return

    cls._pixelLock.acquire()

    # tell VLC to take the data and stuff it into the buffer
    planes[0] = ctypes.cast(cls._framePixelBuffer, ctypes.c_void_p)


@vlc.CallbackDecorators.VideoUnlockCb
def vlcUnlockCallback(user_data, picture, planes):
    """Called when VLC releases the frame draw buffer.
    """
    try:
        cls = ctypes.cast(
            user_data, ctypes.POINTER(ctypes.py_object)).contents.value
    except (ValueError, TypeError):
        return

    cls._pixelLock.release()


@vlc.CallbackDecorators.VideoDisplayCb
def vlcDisplayCallback(user_data, picture):
    """Callback used by VLC when its ready to display a new frame.
    """
    try:
        cls = ctypes.cast(
            user_data, ctypes.POINTER(ctypes.py_object)).contents.value
    except (ValueError, TypeError):
        return

    cls._frameCounter += 1


@vlc.CallbackDecorators.LogCb
def vlcLogCallback(user_data, level, ctx, fmt, args):
    """Callback for logging messages emitted by VLC. Processed here and
    converted to PsychoPy logging format.
    """
    # if level == vlc.DEBUG:
    #     logging.debug()

    pass  # suppress messages from VLC, look scary but can be mostly ignored


def vlcMediaEventCallback(event, user_data, player):
    """Callback used by VLC for handling media player events.
    """
    if not user_data():
        return

    cls = user_data()  # ref to movie class
    event = event.type

    if event == vlc.EventType.MediaPlayerEndReached:
        cls._onEos()


if __name__ == "__main__":
    pass
